# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT Panel widgets (core.gui.panel)

Signal and Image Panel widgets relie on components:

  * `ObjectProp`: widget handling signal/image properties
  using a guidata DataSet

  * `core.gui.panel.objectlist.ObjectList`: widget handling signal/image list

  * `core.gui.panel.actionhandler.SignalActionHandler` or `ImageActionHandler`:
  classes handling Qt actions

  * `core.gui.panel.plotitemlist.SignalItemList` or `ImageItemList`:
  classes handling guiqwt plot items

  * `core.gui.panel.processor.signal.SignalProcessor` or
  `core.gui.panel.processor.image.ImageProcessor`: classes handling computing features

  * `core.gui.panel.roieditor.SignalROIEditor` or `ImageROIEditor`:
  classes handling ROI editor widgets
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

import abc
import dataclasses
import os.path as osp
import re
import warnings
from typing import List

import guidata.dataset.qtwidgets as gdq
import numpy as np
from guidata.configtools import get_icon
from guidata.qthelpers import add_actions
from guidata.utils import update_dataset
from guidata.widgets.arrayeditor import ArrayEditor
from guiqwt.io import imread, imwrite, iohandler
from guiqwt.plot import CurveDialog, ImageDialog
from guiqwt.tools import (
    AnnotatedCircleTool,
    AnnotatedEllipseTool,
    AnnotatedPointTool,
    AnnotatedRectangleTool,
    AnnotatedSegmentTool,
    HCursorTool,
    LabelTool,
    RectangleTool,
    SegmentTool,
    VCursorTool,
    XCursorTool,
)
from qtpy import QtCore as QC
from qtpy import QtWidgets as QW
from qtpy.compat import getopenfilename, getopenfilenames, getsavefilename

from codraft.config import APP_NAME, Conf, _
from codraft.core.gui import actionhandler, objectlist, plotitemlist, roieditor
from codraft.core.gui.processor.image import ImageProcessor
from codraft.core.gui.processor.signal import SignalProcessor
from codraft.core.io.signal import read_signal, write_signal
from codraft.core.model.base import MetadataItem, ObjectItf, ResultShape
from codraft.core.model.image import (
    ImageDatatypes,
    ImageParam,
    create_image,
    create_image_from_param,
    new_image_param,
)
from codraft.core.model.signal import (
    SignalParam,
    create_signal_from_param,
    new_signal_param,
)
from codraft.utils.qthelpers import (
    exec_dialog,
    qt_try_except,
    qt_try_loadsave_file,
    save_restore_stds,
)

#  Registering MetadataItem edit widget
gdq.DataSetEditLayout.register(MetadataItem, gdq.ButtonWidget)


class ObjectProp(QW.QWidget):
    """Object handling panel properties"""

    def __init__(self, panel, paramclass):
        super().__init__(panel)
        self.paramclass = paramclass
        self.properties = gdq.DataSetEditGroupBox(_("Properties"), paramclass)
        self.properties.SIG_APPLY_BUTTON_CLICKED.connect(panel.properties_changed)
        self.properties.setEnabled(False)
        self.add_prop_layout = QW.QHBoxLayout()
        playout = self.properties.edit.layout
        playout.addLayout(
            self.add_prop_layout, playout.rowCount() - 1, 0, 1, 1, QC.Qt.AlignLeft
        )
        hlayout = QW.QHBoxLayout()
        hlayout.addWidget(self.properties)
        vlayout = QW.QVBoxLayout()
        vlayout.addLayout(hlayout)
        vlayout.addStretch()
        self.setLayout(vlayout)

    def add_button(self, button):
        """Add additional button on bottom of properties panel"""
        self.add_prop_layout.addWidget(button)

    def update_properties_from(self, param: ObjectItf = None):
        """Update properties from signal/image dataset"""
        self.properties.setDisabled(param is None)
        if param is None:
            param = self.paramclass()
        self.properties.dataset.set_defaults()
        update_dataset(self.properties.dataset, param)
        self.properties.get()


class BasePanelMeta(type(QW.QSplitter), abc.ABCMeta):
    """Mixed metaclass to avoid conflicts"""


class BasePanel(QW.QSplitter, metaclass=BasePanelMeta):
    """Object handling the item list, the selected item properties and plot"""

    PANEL_STR = ""  # e.g. "Signal Panel"
    PARAMCLASS = SignalParam  # Replaced by the right class in child object
    DIALOGCLASS = CurveDialog  # Idem
    ANNOTATION_TOOLS = (
        LabelTool,
        VCursorTool,
        HCursorTool,
        XCursorTool,
        SegmentTool,
        RectangleTool,
    )
    DIALOGSIZE = (800, 600)
    PREFIX = ""  # e.g. "s"
    OPEN_FILTERS = ""  # Qt file open dialog filters
    H5_PREFIX = ""
    SIG_STATUS_MESSAGE = QC.Signal(str)  # emitted by "qt_try_except" decorator
    SIG_OBJECT_ADDED = QC.Signal()
    SIG_OBJECT_REMOVED = QC.Signal()
    SIG_UPDATE_PLOT_ITEM = QC.Signal(int)  # Update plot item associated to row number
    SIG_UPDATE_PLOT_ITEMS = QC.Signal()  # Update plot items associated to selected rows
    ROIDIALOGOPTIONS = {}
    ROIDIALOGCLASS = roieditor.BaseROIEditor  # Replaced in child object

    @abc.abstractmethod
    def __init__(self, parent, plotwidget, toolbar):
        super().__init__(QC.Qt.Vertical, parent)
        self.setObjectName(self.PREFIX)
        self.mainwindow = parent
        self.objprop = ObjectProp(self, self.PARAMCLASS)
        self.objlist = objectlist.ObjectList(self)
        self.itmlist = None
        self.processor = None
        self.acthandler = None
        self.__metadata_clipboard = {}
        self.context_menu = QW.QMenu()
        self.__separate_views = {}

    def setup_panel(self):
        """Setup panel"""
        self.processor.SIG_ADD_SHAPE.connect(self.itmlist.add_shapes)
        self.SIG_UPDATE_PLOT_ITEM.connect(self.itmlist.refresh_plot)
        self.SIG_UPDATE_PLOT_ITEMS.connect(self.itmlist.refresh_plot)
        self.objlist.itemSelectionChanged.connect(self.selection_changed)
        self.objlist.SIG_ITEM_DOUBLECLICKED.connect(
            lambda row: self.open_separate_view([row])
        )
        self.objlist.SIG_CONTEXT_MENU.connect(self.__popup_contextmenu)
        self.objprop.properties.SIG_APPLY_BUTTON_CLICKED.connect(
            self.properties_changed
        )
        self.addWidget(self.objlist)
        self.addWidget(self.objprop)
        self.add_results_button()

    def get_category_actions(self, category):  # pragma: no cover
        """Return actions for category"""
        return self.acthandler.feature_actions[category]

    def __popup_contextmenu(self, position: QC.QPoint):  # pragma: no cover
        """Popup context menu at position"""
        # Note: For now, this is completely unnecessary to clear context menu everytime,
        # but implementing it this way could be useful in the future in menu contents
        # should take into account current object selection
        self.context_menu.clear()
        add_actions(self.context_menu, self.acthandler.actlist_cmenu)
        self.context_menu.popup(position)

    # ------Creating, adding, removing objects------------------------------------------
    def create_object(self, title=None):
        """Create object (signal or image)

        :param str title: Title of the object
        """
        obj = self.PARAMCLASS(title=title)
        obj.title = title
        return obj

    @qt_try_except()
    def add_object(self, obj, refresh=True):
        """Add signal/image object and return associated plot item"""
        obj.check_data()
        self.objlist.append(obj)
        item = self.itmlist.append(None)
        if refresh:
            self.objlist.refresh_list(-1)
        self.SIG_OBJECT_ADDED.emit()
        return item

    # TODO: [P2] New feature: move objects up/down
    @qt_try_except()
    def insert_object(self, obj, row, refresh=True):
        """Insert signal/image object after row"""
        obj.check_data()
        self.objlist.insert(row, obj)
        self.itmlist.insert(row)
        if refresh:
            self.objlist.refresh_list(new_current_row=row + 1)
        self.SIG_OBJECT_ADDED.emit()

    def duplicate_object(self):
        """Duplication signal/image object"""
        if not self.mainwindow.confirm_memory_state():
            return
        rows = sorted(self.objlist.get_selected_rows())
        row = None
        for row in rows:
            obj = self.objlist[row]
            objcopy = self.create_object()
            objcopy.title = obj.title
            objcopy.copy_data_from(obj)
            self.add_object(objcopy, refresh=False)
        self.objlist.refresh_list(new_current_row=-1)
        self.SIG_UPDATE_PLOT_ITEMS.emit()

    def copy_metadata(self):
        """Copy object metadata"""
        row = self.objlist.get_selected_rows()[0]
        obj = self.objlist[row]
        self.__metadata_clipboard = obj.metadata.copy()
        pfx = self.objlist.prefix
        new_pref = f"{pfx}{row:03d}_"
        for key, value in obj.metadata.items():
            if ResultShape.match(key, value):
                mshape = ResultShape.from_metadata_entry(key, value)
                if not re.match(pfx + r"[0-9]{3}[\s]*", mshape.label):
                    # Handling additional result (e.g. diameter)
                    for a_key, a_value in obj.metadata.items():
                        if isinstance(a_key, str) and a_key.startswith(mshape.label):
                            self.__metadata_clipboard.pop(a_key)
                            self.__metadata_clipboard[new_pref + a_key] = a_value
                    mshape.label = new_pref + mshape.label
                    # Handling result shape
                    self.__metadata_clipboard.pop(key)
                    self.__metadata_clipboard[mshape.key] = value

    def paste_metadata(self):
        """Paste metadata to selected object(s)"""
        rows = sorted(self.objlist.get_selected_rows(), reverse=True)
        row = None
        for row in rows:
            obj = self.objlist[row]
            obj.metadata.update(self.__metadata_clipboard)
        self.SIG_UPDATE_PLOT_ITEMS.emit()

    def remove_object(self):
        """Remove signal/image object"""
        rows = sorted(self.objlist.get_selected_rows(), reverse=True)
        for row in rows:
            for dlg, obj in self.__separate_views.items():
                if obj is self.objlist[row]:
                    dlg.done(QW.QDialog.DialogCode.Rejected)
            del self.objlist[row]
            del self.itmlist[row]
        self.objlist.refresh_list(max(0, rows[-1] - 1))
        self.SIG_UPDATE_PLOT_ITEMS.emit()
        self.SIG_OBJECT_REMOVED.emit()

    def delete_all_objects(self):  # pragma: no cover
        """Confirm before removing all objects"""
        if len(self.objlist) == 0:
            return
        answer = QW.QMessageBox.warning(
            self,
            _("Delete all"),
            _("Do you want to delete all objects (%s)?") % self.PANEL_STR,
            QW.QMessageBox.Yes | QW.QMessageBox.No,
        )
        if answer == QW.QMessageBox.Yes:
            self.remove_all_objects()

    def remove_all_objects(self):
        """Remove all signal/image objects"""
        for dlg in self.__separate_views:
            dlg.done(QW.QDialog.DialogCode.Rejected)
        self.objlist.remove_all()
        self.itmlist.remove_all()
        self.objlist.refresh_list(0)
        self.SIG_UPDATE_PLOT_ITEMS.emit()
        self.SIG_OBJECT_REMOVED.emit()

    def delete_metadata(self):
        """Delete object metadata"""
        for index, row in enumerate(self.objlist.get_selected_rows()):
            self.objlist[row].metadata = {}
            if index == 0:
                self.selection_changed()
        self.SIG_UPDATE_PLOT_ITEMS.emit()

    @abc.abstractmethod
    def new_object(self, newparam=None, addparam=None, edit=True):
        """Create a new object (signal/image).

        :param guidata.dataset.DataSet newparam: new object parameters
        :param guidata.dataset.datatypes.DataSet addparam: additional parameters
        :param bool edit: Open a dialog box to edit parameters (default: True)
        """

    @abc.abstractmethod
    def open_object(self, filename: str) -> None:
        """Open object from file (signal/image)"""

    def open_objects(self, filenames: List[str] = None) -> None:
        """Open objects from file (signals/images)"""
        if not self.mainwindow.confirm_memory_state():
            return
        if filenames is None:  # pragma: no cover
            basedir = Conf.main.base_dir.get()
            with save_restore_stds():
                filenames, _filter = getopenfilenames(
                    self, _("Open"), basedir, self.OPEN_FILTERS
                )
        for filename in filenames:
            with qt_try_loadsave_file(self.parent(), filename, "load"):
                Conf.main.base_dir.set(filename)
                self.open_object(filename)

    def save_objects(self, filenames: List[str] = None) -> None:
        """Save selected objects to file (signal/image)"""
        rows = self.objlist.get_selected_rows()
        if filenames is None:  # pragma: no cover
            filenames = [None] * len(rows)
        assert len(filenames) == len(rows)
        for index, row in enumerate(rows):
            filename = filenames[index]
            obj = self.objlist[row]
            self.save_object(obj, filename)

    @abc.abstractmethod
    def save_object(self, obj, filename: str = None) -> None:
        """Save object to file (signal/image)"""

    def import_metadata_from_file(self, filename: str = None):
        """Import metadata from file (JSON)"""
        if filename is None:  # pragma: no cover
            basedir = Conf.main.base_dir.get()
            with save_restore_stds():
                filename, _filter = getopenfilename(
                    self, _("Import metadata"), basedir, "*.json"
                )
        if filename:
            with qt_try_loadsave_file(self.parent(), filename, "load"):
                Conf.main.base_dir.set(filename)
                row = self.objlist.get_selected_rows()[0]
                obj = self.objlist[row]
                obj.import_metadata_from_file(filename)
            self.SIG_UPDATE_PLOT_ITEMS.emit()

    def export_metadata_from_file(self, filename: str = None):
        """Export metadata to file (JSON)"""
        row = self.objlist.get_selected_rows()[0]
        obj = self.objlist[row]
        if filename is None:  # pragma: no cover
            basedir = Conf.main.base_dir.get()
            with save_restore_stds():
                filename, _filt = getsavefilename(
                    self, _("Export metadata"), basedir, "*.json"
                )
        if filename:
            with qt_try_loadsave_file(self.parent(), filename, "save"):
                Conf.main.base_dir.set(filename)
                obj.export_metadata_to_file(filename)

    # ------Serializing/deserializing objects-------------------------------------------
    def serialize_to_hdf5(self, writer):
        """Serialize objects to a HDF5 file"""
        with writer.group(self.H5_PREFIX):
            for idx, obj in enumerate(self.objlist):
                title = re.sub("[^-a-zA-Z0-9_.() ]+", "", obj.title.replace("/", "_"))
                name = f"{self.PREFIX}{idx:03d}: {title}"
                with writer.group(name):
                    obj.serialize(writer)

    def deserialize_from_hdf5(self, reader):
        """Deserialize objects from a HDF5 file"""
        with reader.group(self.H5_PREFIX):
            for name in reader.h5.get(self.H5_PREFIX, []):
                obj = self.PARAMCLASS()
                with reader.group(name):
                    obj.deserialize(reader)
                    self.add_object(obj)
                    QW.QApplication.processEvents()

    # ------Refreshing GUI--------------------------------------------------------------
    def selection_changed(self):
        """Signal list: selection changed"""
        row = self.objlist.currentRow()
        sel_objs = self.objlist.get_sel_objects()
        if not sel_objs:
            row = -1
        self.objprop.update_properties_from(self.objlist[row] if row != -1 else None)
        self.SIG_UPDATE_PLOT_ITEMS.emit()
        self.acthandler.selection_rows_changed()

    def properties_changed(self):
        """The properties 'Apply' button was clicked: updating signal"""
        row = self.objlist.currentRow()
        update_dataset(self.objlist[row], self.objprop.properties.dataset)
        self.objlist.refresh_list()
        self.SIG_UPDATE_PLOT_ITEMS.emit()

    # ------Plotting data in modal dialogs----------------------------------------------
    def open_separate_view(self, rows=None) -> QW.QDialog:
        """
        Open separate view for visualizing selected objects

        :param list rows: List of row indexes for the objects to be shown in dialog
        :return: Dialog instance
        """
        title = _("Annotations")
        if rows is None:
            rows = self.objlist.get_selected_rows()
        row = rows[0]
        obj = self.objlist[row]
        dlg = self.create_new_dialog(rows, edit=True, name="new_window")
        width, height = self.DIALOGSIZE
        dlg.resize(width, height)
        dlg.plot_widget.itemlist.setVisible(True)
        toolbar = QW.QToolBar(title, self)
        dlg.button_layout.insertWidget(0, toolbar)
        # dlg.layout().insertWidget(1, toolbar)  # other possible location
        # dlg.plot_layout.addWidget(toolbar, 1, 0, 1, 1)  # other possible location
        dlg.add_toolbar(toolbar, id(toolbar))
        toolbar.setToolButtonStyle(QC.Qt.ToolButtonTextUnderIcon)
        for tool in self.ANNOTATION_TOOLS:
            dlg.add_tool(tool, toolbar_id=id(toolbar))
        plot = dlg.get_plot()
        plot.unselect_all()
        for item in plot.items:
            item.set_selectable(False)
        for item in obj.iterate_shape_items(editable=True):
            plot.add_item(item)
        self.__separate_views[dlg] = obj
        dlg.show()
        dlg.finished.connect(self.__separate_view_finished)
        return dlg

    def __separate_view_finished(self, result: int):
        """Separate view was closed"""
        dlg = self.sender()
        if result == QW.QDialog.DialogCode.Accepted:
            items = dlg.get_plot().get_items()
            rw_items = [item for item in items if not item.is_readonly()]
            if rw_items:
                obj = self.__separate_views[dlg]
                obj.set_annotations_from_items(rw_items)
                self.selection_changed()
                self.SIG_UPDATE_PLOT_ITEMS.emit()

    def toggle_show_titles(self, state):
        """Toggle show annotations option"""
        Conf.view.show_label.set(state)
        for obj in self.objlist:
            obj.metadata[obj.METADATA_LBL] = state
        self.SIG_UPDATE_PLOT_ITEMS.emit()

    def create_new_dialog(
        self,
        rows,
        edit=False,
        toolbar=True,
        title=None,
        tools=None,
        name=None,
        options=None,
    ):
        """
        Create new pop-up signal/image plot dialog

        :param list rows: List of row indexes for the objects to be shown in dialog
        :param bool edit: If True, show "OK" and "Cancel" buttons
        :param bool toolbar: If True, add toolbar
        :param str title: Title of the dialog box
        :param list tools: List of plot tools
        :param str name: Name of the widget (used as screenshot basename)
        :param dict options: Plot options
        """
        if title is not None or len(rows) == 1:
            if title is None:
                title = self.objlist.get_sel_object().title
            title = f"{title} - {APP_NAME}"
        else:
            title = APP_NAME
        plot_options = self.itmlist.get_current_plot_options()
        if options is not None:
            plot_options.update(options)
        dlg = self.DIALOGCLASS(
            parent=self,
            wintitle=title,
            edit=edit,
            options=plot_options,
            toolbar=toolbar,
        )
        dlg.setWindowIcon(get_icon("codraft.svg"))
        dlg.setObjectName(f"{self.PREFIX}_{name}")
        if tools is not None:
            for tool in tools:
                dlg.add_tool(tool)
        plot = dlg.get_plot()
        for row in rows:
            item = self.itmlist.make_item_from_existing(row)
            item.set_readonly(True)
            plot.add_item(item, z=0)
        plot.set_active_item(item)
        plot.replot()
        return dlg

    def create_new_dialog_for_selection(
        self, title, name, options=None, toolbar=False, tools=None
    ):
        """
        Create new pop-up dialog for the currently selected signal/image

        :param str title: Title of the dialog box
        :param str name: Name of the widget (used as screenshot basename)
        :param dict options: Plot options
        :param list tools: List of plot tools
        :return: tuple (dialog, current_object)
        """
        row = self.objlist.get_selected_rows()[0]
        obj = self.objlist[row]
        dlg = self.create_new_dialog(
            [row],
            edit=True,
            toolbar=toolbar,
            title=f"{title} - {obj.title}",
            tools=tools,
            name=name,
            options=options,
        )
        return dlg, obj

    def get_roi_dialog(self, extract: bool, singleobj: bool) -> roieditor.ROIEditorData:
        """Get ROI data (array) from specific dialog box"""
        roi_s = _("Regions of interest")
        options = self.ROIDIALOGOPTIONS
        dlg, obj = self.create_new_dialog_for_selection(roi_s, "roi_dialog", options)
        plot = dlg.get_plot()
        plot.unselect_all()
        for item in plot.items:
            item.set_selectable(False)
        roi_editor = self.ROIDIALOGCLASS(dlg, obj, extract, singleobj)
        dlg.plot_layout.addWidget(roi_editor, 1, 0, 1, 1)
        if exec_dialog(dlg):
            return roi_editor.get_data()
        return None

    def get_object_dialog(
        self, parent: QW.QWidget, title: str
    ) -> objectlist.GetObjectDialog:
        """Get object dialog"""
        dlg = objectlist.GetObjectDialog(parent, self, title)
        if exec_dialog(dlg):
            return dlg.get_object()
        return None

    def add_results_button(self):
        """Add 'Show results' button"""
        btn = QW.QPushButton(get_icon("show_results.svg"), _("Show results"), self)
        btn.setToolTip(_("Show results obtained from previous computations"))
        self.objprop.add_button(btn)
        btn.clicked.connect(self.show_results)
        self.acthandler.actlist_1more.append(btn)

    def show_results(self):
        """Show results"""
        rows = self.objlist.get_selected_rows()

        @dataclasses.dataclass
        class ResultData:
            """Result data associated to a shapetype"""

            results: List[ResultShape] = None
            xlabels: List[str] = None
            ylabels: List[str] = None

        rdatadict = {}
        for idx, row in enumerate(rows):
            obj = self.objlist[row]
            for key, value in obj.metadata.items():
                if ResultShape.match(key, value):
                    result = ResultShape.from_metadata_entry(key, value)
                    rdata = rdatadict.setdefault(
                        result.shapetype, ResultData([], None, [])
                    )
                    title = f"{result.label}"
                    rdata.results.append(result)
                    rdata.xlabels = result.xlabels
                    for _i_row_res in range(result.array.shape[0]):
                        ylabel = f"{self.PREFIX}{idx:03d}: {result.label}"
                        rdata.ylabels.append(ylabel)
        if rdatadict:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", RuntimeWarning)
                for rdata in rdatadict.values():
                    dlg = ArrayEditor(self.parent())
                    title = _("Results")
                    dlg.setup_and_check(
                        np.vstack([result.array for result in rdata.results]),
                        title,
                        readonly=True,
                        xlabels=rdata.xlabels,
                        ylabels=rdata.ylabels,
                    )
                    dlg.setObjectName(f"{self.PREFIX}_results")
                    dlg.resize(750, 300)
                    exec_dialog(dlg)
        else:
            msg = "<br>".join(
                [
                    _("No result currently available for this object."),
                    "",
                    _(
                        "This feature shows result arrays as displayed after "
                        'calling one of the computing feature (see "Compute" menu).'
                    ),
                ]
            )
            QW.QMessageBox.information(self, APP_NAME, msg)


class SignalPanel(BasePanel):
    """Object handling the item list, the selected item properties and plot,
    specialized for Signal objects"""

    PANEL_STR = _("Signal List")
    PARAMCLASS = SignalParam
    DIALOGCLASS = CurveDialog
    PREFIX = "s"
    OPEN_FILTERS = f'{_("Text files")} (*.txt *.csv)\n{_("NumPy arrays")} (*.npy)'
    H5_PREFIX = "CodraFT_Sig"
    ROIDIALOGCLASS = roieditor.SignalROIEditor

    # pylint: disable=duplicate-code

    def __init__(self, parent, plotwidget, toolbar):
        super().__init__(parent, plotwidget, toolbar)
        self.itmlist = plotitemlist.SignalItemList(self, self.objlist, plotwidget)
        self.processor = SignalProcessor(self, self.objlist, plotwidget)
        self.acthandler = actionhandler.SignalActionHandler(
            self, self.objlist, self.itmlist, self.processor, toolbar
        )
        self.setup_panel()

    # ------Creating, adding, removing objects------------------------------------------
    def new_object(self, newparam=None, addparam=None, edit=True):
        """Create a new signal.

        :param codraft.core.model.signal.SignalNewParam newparam: new signal parameters
        :param guidata.dataset.datatypes.DataSet addparam: additional parameters
        :param bool edit: Open a dialog box to edit parameters (default: True)
        """
        if not self.mainwindow.confirm_memory_state():
            return
        curobj = self.objlist.get_sel_object(-1)
        if curobj is not None:
            newparam = newparam if newparam is not None else new_signal_param()
            newparam.size = len(curobj.data)
            newparam.xmin = curobj.x.min()
            newparam.xmax = curobj.x.max()
        signal = create_signal_from_param(
            newparam, addparam=addparam, edit=edit, parent=self
        )
        if signal is not None:
            self.add_object(signal)

    def open_object(self, filename: str) -> None:
        """Open object from file (signal/image)"""
        signal = read_signal(filename)
        self.add_object(signal)

    def save_object(self, obj, filename: str = None) -> None:
        """Save object to file (signal/image)"""
        if filename is None:
            basedir = Conf.main.base_dir.get()
            with save_restore_stds():
                filename, _filter = getsavefilename(  # pylint: disable=duplicate-code
                    self, _("Save as"), basedir, self.OPEN_FILTERS
                )
        if filename:
            with qt_try_loadsave_file(self.parent(), filename, "save"):
                Conf.main.base_dir.set(filename)
                write_signal(obj, filename)


class ImagePanel(BasePanel):
    """Object handling the item list, the selected item properties and plot,
    specialized for Image objects"""

    PANEL_STR = _("Image List")
    PARAMCLASS = ImageParam
    DIALOGCLASS = ImageDialog
    DIALOGSIZE = (800, 800)
    ANNOTATION_TOOLS = (
        AnnotatedCircleTool,
        AnnotatedSegmentTool,
        AnnotatedRectangleTool,
        AnnotatedPointTool,
        AnnotatedEllipseTool,
        LabelTool,
    )
    PREFIX = "i"
    OPEN_FILTERS = iohandler.get_filters("load", dtype=None)
    H5_PREFIX = "CodraFT_Ima"
    ROIDIALOGOPTIONS = dict(show_itemlist=True, show_contrast=False)
    ROIDIALOGCLASS = roieditor.ImageROIEditor

    # pylint: disable=duplicate-code

    def __init__(self, parent, plotwidget, toolbar):
        super().__init__(parent, plotwidget, toolbar)
        self.itmlist = plotitemlist.ImageItemList(self, self.objlist, plotwidget)
        self.processor = ImageProcessor(self, self.objlist, plotwidget)
        self.acthandler = actionhandler.ImageActionHandler(
            self, self.objlist, self.itmlist, self.processor, toolbar
        )
        self.setup_panel()

    # ------Refreshing GUI--------------------------------------------------------------
    def properties_changed(self):
        """The properties 'Apply' button was clicked: updating signal"""
        row = self.objlist.currentRow()
        self.objlist[row].invalidate_maskdata_cache()
        super().properties_changed()

    # ------Creating, adding, removing objects------------------------------------------
    def new_object(self, newparam=None, addparam=None, edit=True):
        """Create a new image.

        :param codraft.core.model.image.ImageNewParam newparam: new image parameters
        :param guidata.dataset.datatypes.DataSet addparam: additional parameters
        :param bool edit: Open a dialog box to edit parameters (default: True)
        """
        if not self.mainwindow.confirm_memory_state():
            return
        curobj = self.objlist.get_sel_object(-1)
        if curobj is not None:
            newparam = newparam if newparam is not None else new_image_param()
            newparam.width, newparam.height = curobj.size
            newparam.dtype = ImageDatatypes.from_dtype(curobj.data.dtype)
        image = create_image_from_param(
            newparam, addparam=addparam, edit=edit, parent=self
        )
        if image is not None:
            self.add_object(image)

    def open_object(self, filename: str) -> None:
        """Open object from file (signal/image)"""
        data = imread(filename, to_grayscale=False)
        reducepath = osp.relpath(filename, osp.join(osp.dirname(filename), osp.pardir))
        if filename.lower().endswith(".sif") and len(data.shape) == 3:
            for idx in range(data.shape[0]):
                image = create_image(reducepath + "_Im" + str(idx), data[idx, ::])
                self.add_object(image)
        else:
            if data.ndim == 3:
                # Converting to grayscale
                data = data[..., :4].mean(axis=2)
            image = create_image(reducepath, data)
            if osp.splitext(filename)[1].lower() == ".dcm":
                from pydicom import dicomio  # pylint: disable=C0415,E0401

                image.dicom_template = dicomio.read_file(
                    filename, stop_before_pixels=True, force=True
                )
            self.add_object(image)

    def save_object(self, obj, filename: str = None) -> None:
        """Save object to file (signal/image)"""
        if filename is None:
            basedir = Conf.main.base_dir.get()
            with save_restore_stds():
                filename, _filter = getsavefilename(  # pylint: disable=duplicate-code
                    self,
                    _("Save as"),
                    basedir,
                    iohandler.get_filters(
                        "save", dtype=obj.data.dtype, template=obj.dicom_template
                    ),
                )
        if filename:
            kwargs = {}
            if osp.splitext(filename)[1].lower() == ".dcm":
                kwargs["template"] = obj.dicom_template
            with qt_try_loadsave_file(self.parent(), filename, "save"):
                Conf.main.base_dir.set(filename)
                imwrite(filename, obj.data, **kwargs)

    def toggle_show_contrast(self, state):
        """Toggle show contrast option"""
        Conf.view.show_contrast.set(state)
        self.SIG_UPDATE_PLOT_ITEMS.emit()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Signal peak detection test

Testing peak detection dialog box.
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...


from codraft.core.io.signal import read_signal
from codraft.env import execenv
from codraft.utils.qthelpers import exec_dialog, qt_app_context
from codraft.utils.tests import get_test_fnames
from codraft.widgets.signalpeakdialog import SignalPeakDetectionDialog

SHOW = True  # Show test in GUI-based test launcher


def test():
    """Signal peak dialog test"""
    with qt_app_context():
        s = read_signal(get_test_fnames("paracetamol.txt")[0])
        dlg = SignalPeakDetectionDialog()
        dlg.resize(640, 300)
        dlg.setup_data(s.x, s.y)
        plot = dlg.get_plot()
        plot.set_axis_limits(plot.xBottom, 16, 30)
        dlg.setObjectName(dlg.objectName() + "_00")  # to avoid timestamp suffix
        exec_dialog(dlg)
    execenv.print("peaks:")
    execenv.pprint(dlg.get_peaks())
    execenv.pprint(dlg.get_min_dist())


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
I/O application test:

  - Testing signals I/O
  - Testing images I/O
"""

import os.path as osp

from guiqwt.io import iohandler

from codraft.env import execenv
from codraft.tests import codraft_app_context
from codraft.utils.tests import get_test_fnames, temporary_directory

SHOW = True  # Show test in GUI-based test launcher


def __test_io_features(title, panel, pattern, getreadfunc=None, getwritefunc=None):
    """ "Test I/O features"""
    execenv.print(f"  {title}:")
    with temporary_directory() as tmpdir:
        # os.startfile(tmpdir)
        fnames = get_test_fnames(pattern)
        execenv.print("    Opening:")
        for fname in fnames:
            if getreadfunc is not None:
                try:
                    getreadfunc(osp.splitext(fname)[1])
                except RuntimeError:
                    execenv.print(f"      Skipping {fname}")
                    continue
            execenv.print(f"      {fname}")
            panel.open_object(fname)
        execenv.print("    Saving:")
        for row, fname in enumerate(fnames):
            panel.objlist.set_current_row(row)
            if getwritefunc is not None:
                try:
                    getwritefunc(osp.splitext(fname)[1])
                except RuntimeError:
                    execenv.print(f"      Skipping {fname}")
                    continue
            path = osp.join(tmpdir, osp.basename(fname))
            execenv.print(f"      {path}")
            panel.save_objects([path])


def test():
    """Run image tools test scenario"""
    with codraft_app_context() as win:
        execenv.print("I/O application test:")

        # === Testing Signal I/O ---------------------------------------------------
        __test_io_features("Signals", win.signalpanel, "curve_formats/*.*")

        # === Testing Image I/O ----------------------------------------------------
        __test_io_features(
            "Images",
            win.imagepanel,
            "image_formats/*.*",
            getreadfunc=iohandler.get_readfunc,
            getwritefunc=iohandler.get_writefunc,
        )


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Copyright © 2022 Codra
# Pierre Raybaut

"""
Module providing CodraFT Installation configuration widget
"""

import locale
import os
import platform
import sys
from subprocess import PIPE, Popen

from guidata.configtools import get_icon, get_module_data_path
from guidata.widgets.codeeditor import CodeEditor
from qtpy import QtWidgets as QW

from codraft import __version__
from codraft.config import _
from codraft.utils import dephash
from codraft.utils.qthelpers import exec_dialog


def decode_fs_string(string):
    """Convert string from file system charset to unicode"""
    charset = sys.getfilesystemencoding()
    if charset is None:
        charset = locale.getpreferredencoding()
    return string.decode(charset)


def get_pip_list():
    """Get pip list result"""
    command = " ".join([sys.executable, "-m", "pip list"])
    with Popen(command, shell=True, stdout=PIPE, stderr=PIPE) as proc:
        output = proc.stdout.read()
    return decode_fs_string(output)


class InstallConfigViewerWidget(QW.QWidget):
    """Installation configuration widget"""

    def __init__(self, label, contents, parent=None):
        super().__init__(parent)
        self.editor = CodeEditor()
        self.editor.setReadOnly(True)
        self.editor.setPlainText(contents)
        layout = QW.QVBoxLayout()
        layout.addWidget(QW.QLabel(label))
        layout.addWidget(self.editor)
        self.setLayout(layout)


class InstallConfigViewerWindow(QW.QDialog):
    """Installation configuration window"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setObjectName("dependencies")
        self.setWindowTitle(_("About CodraFT installation"))
        self.setWindowIcon(get_icon("codraft.svg"))
        self.tabs = QW.QTabWidget()
        label = "Information on current CodraFT installation:"
        more_infos = ""
        try:
            datapath = get_module_data_path("codraft", "data")
            state = dephash.check_dependencies_hash(datapath)
            bad_deps = [name for name in state if not state[name]]
            if bad_deps:
                more_infos += "Invalid dependencies: "
                more_infos += ", ".join(bad_deps)
            else:
                more_infos += "Dependencies hash file: checked."
        except IOError:
            more_infos += "Unable to open dependencies hash file."
        more_infos += os.linesep * 2
        if sys.executable.lower().endswith("codraft.exe"):
            #  Stand-alone version
            more_infos += "This is the Stand-alone version of CodraFT"
        else:
            more_infos += get_pip_list()
        infos = os.linesep.join(
            [
                f"CodraFT v{__version__}",
                "",
                f"Machine type: {platform.machine()}",
                f"Platform: {platform.platform()}",
                f"Python v{sys.version}",
                "",
                more_infos,
            ]
        )
        viewer = InstallConfigViewerWidget(label, infos, parent=self)
        layout = QW.QVBoxLayout()
        layout.addWidget(viewer)
        self.setLayout(layout)
        self.resize(800, 500)


def exec_codraft_installconfig_dialog(parent=None):
    """View CodraFT installation configuration"""
    dlg = InstallConfigViewerWindow(parent=parent)
    exec_dialog(dlg)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT Demo
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code

from qtpy import QtWidgets as QW

from codraft.config import _, reset
from codraft.core.gui.main import CodraFTMainWindow
from codraft.core.gui.processor.image import ContourShapeParam
from codraft.core.gui.processor.image import PeakDetectionParam as Peak2DDetectionParam
from codraft.core.gui.processor.image import RotateParam
from codraft.core.gui.processor.signal import PeakDetectionParam, PolynomialFitParam
from codraft.core.model.base import UniformRandomParam
from codraft.core.model.image import ImageTypes, create_image, new_image_param
from codraft.core.model.signal import (
    GaussLorentzVoigtParam,
    SignalParamNew,
    SignalTypes,
    create_signal_from_param,
    new_signal_param,
)
from codraft.env import execenv
from codraft.tests import codraft_app_context
from codraft.tests.data import (
    PeakDataParam,
    create_test_image1,
    create_test_image2,
    create_test_signal1,
    get_peak2d_data,
)
from codraft.tests.roi_app import create_test_image_with_roi
from codraft.tests.scenario_sig_app import test_common_operations
from codraft.utils.qthelpers import qt_wait
from codraft.widgets import fitdialog

DELAY1, DELAY2, DELAY3 = 1, 2, 4
# DELAY1, DELAY2, DELAY3 = 0, 0, 0


def test_signal_features(win: CodraFTMainWindow, data_size: int = 500) -> None:
    """Testing signal features"""
    panel = win.signalpanel
    win.switch_to_signal_panel()

    qt_wait(DELAY2)

    sig1 = create_test_signal1(data_size)
    win.add_object(sig1)

    panel.objlist.set_current_row(0)
    newparam = new_signal_param(_("Random function"), stype=SignalTypes.UNIFORMRANDOM)
    addparam = UniformRandomParam()
    addparam.vmin = 0
    addparam.vmax = sig1.y.max() * 0.2
    panel.new_object(newparam, addparam=addparam, edit=True)

    test_common_operations(panel)
    panel.processor.normalize()
    panel.processor.compute_derivative()
    panel.processor.compute_integral()

    param = PeakDetectionParam()
    panel.processor.detect_peaks(param)

    qt_wait(DELAY2)

    param = PolynomialFitParam()
    panel.processor.compute_polyfit(param)

    panel.processor.compute_fit(_("Gaussian fit"), fitdialog.gaussianfit)

    newparam = SignalParamNew()
    newparam.title = _("Gaussian")
    newparam.type = SignalTypes.GAUSS
    sig = create_signal_from_param(newparam, GaussLorentzVoigtParam(), edit=False)
    panel.add_object(sig)

    panel.processor.compute_fwhm()
    panel.processor.compute_fw1e2()

    qt_wait(DELAY2)


def test_image_features(win: CodraFTMainWindow, data_size: int = 1000) -> None:
    """Testing signal features"""
    win.switch_to_image_panel()
    panel = win.imagepanel

    ima1 = create_test_image2(data_size)
    panel.add_object(ima1)

    qt_wait(DELAY2)

    panel.objlist.set_current_row(0)
    newparam = new_image_param(itype=ImageTypes.UNIFORMRANDOM)
    addparam = UniformRandomParam()
    addparam.set_from_datatype(ima1.data.dtype)
    addparam.vmax = int(ima1.data.max() * 0.2)
    panel.new_object(newparam, addparam=addparam, edit=False)

    test_common_operations(panel)

    ima1 = create_test_image1(data_size)
    panel.add_object(ima1)

    qt_wait(DELAY3)

    panel.processor.rotate_90()
    panel.processor.rotate_270()
    panel.processor.flip_horizontally()
    panel.processor.flip_vertically()

    param = RotateParam()
    param.angle = 5.0
    for boundary in param.boundaries[:-1]:
        param.mode = boundary
        panel.processor.rotate_arbitrarily(param)

    ima1 = create_test_image_with_roi(data_size)
    panel.add_object(ima1)

    qt_wait(DELAY2)

    panel.processor.compute_centroid()

    qt_wait(DELAY1)

    panel.processor.compute_enclosing_circle()

    qt_wait(DELAY2)

    data = get_peak2d_data(PeakDataParam(size=data_size))
    ima = create_image("Test image with peaks", data)
    panel.add_object(ima)
    param = Peak2DDetectionParam()
    param.create_rois = True
    panel.processor.compute_peak_detection(param)

    qt_wait(DELAY3)

    param = ContourShapeParam()
    panel.processor.compute_contour_shape(param)

    qt_wait(DELAY3)

    panel.processor.extract_roi()


def run():
    """Run demo"""
    reset()  # Reset configuration (remove configuration file and initialize it)
    execenv.enable_demo_mode(DELAY1)
    with codraft_app_context(console=False) as win:
        QW.QMessageBox.information(win, "Demo", "Click OK to start demo")
        test_signal_features(win)
        test_image_features(win)
        qt_wait(DELAY3)
        QW.QMessageBox.information(win, "Demo", "Click OK to end demo")


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT Qt utilities
"""

import faulthandler
import functools
import logging
import os
import os.path as osp
import shutil
import sys
import time
import traceback
from contextlib import contextmanager
from datetime import datetime

import guidata
from guidata.configtools import get_module_data_path
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW

from codraft.config import APP_NAME, DATETIME_FORMAT, Conf, _, get_old_log_fname
from codraft.env import execenv
from codraft.utils.misc import to_string


def close_widgets_and_quit(screenshot=False):
    """Close Qt top level widgets and quit Qt event loop"""
    for widget in QW.QApplication.instance().topLevelWidgets():
        wname = widget.objectName()
        if screenshot and wname and widget.isVisible():  # pragma: no cover
            grab_save_window(widget, wname.lower())
        assert widget.close()
    QW.QApplication.instance().quit()


QAPP_INSTANCE = None


def get_log_contents(fname):
    """Return True if file exists and something was logged in it"""
    if osp.exists(fname):
        with open(fname, "rb") as fdesc:
            return to_string(fdesc.read()).strip()
    return None


def initialize_log_file(fname):
    """Eventually keep the previous log file
    Returns True if there was a previous log file"""
    contents = get_log_contents(fname)
    if contents:
        try:
            shutil.move(fname, get_old_log_fname(fname))
        except Exception:  # pylint: disable=broad-except
            pass
        return True
    return False


def remove_empty_log_file(fname):
    """Eventually remove empty log files"""
    if not get_log_contents(fname):
        try:
            os.remove(fname)
        except Exception:  # pylint: disable=broad-except
            pass


@contextmanager
def qt_app_context(exec_loop=False, enable_logs=True):
    """Context manager handling Qt application creation and persistance"""
    global QAPP_INSTANCE  # pylint: disable=global-statement
    if QAPP_INSTANCE is None:
        QAPP_INSTANCE = guidata.qapplication()

    if enable_logs:
        # === Create a logger for standard exceptions ----------------------------------
        tb_log_fname = Conf.main.traceback_log_path.get()
        Conf.main.traceback_log_available.set(initialize_log_file(tb_log_fname))
        logger = logging.getLogger(__name__)
        fmt = "[%(asctime)s] {%(pathname)s:%(lineno)d} %(levelname)s - %(message)s"
        logging.basicConfig(
            filename=tb_log_fname,
            filemode="w",
            level=logging.CRITICAL,
            format=fmt,
            datefmt=DATETIME_FORMAT,
        )

        def custom_excepthook(exc_type, exc_value, exc_traceback):
            "Custom exception hook"
            logger.critical(
                "Unhandled exception", exc_info=(exc_type, exc_value, exc_traceback)
            )
            return sys.__excepthook__(exc_type, exc_value, exc_traceback)

        sys.excepthook = custom_excepthook

    # === Use faulthandler for other exceptions ------------------------------------
    fh_log_fname = Conf.main.faulthandler_log_path.get()
    Conf.main.faulthandler_log_available.set(initialize_log_file(fh_log_fname))

    with open(fh_log_fname, "w", encoding="utf-8") as fh_log_fn:
        if enable_logs and Conf.main.faulthandler_enabled.get(True):
            faulthandler.enable(file=fh_log_fn)
        try:
            yield QAPP_INSTANCE
        finally:
            if execenv.unattended:  # pragma: no cover
                if execenv.delay > 0:
                    mode = "Screenshot" if execenv.screenshot else "Unattended"
                    message = f"{mode} mode (delay: {execenv.delay}s)"
                    msec = execenv.delay * 1000 - 200
                    for widget in QW.QApplication.instance().topLevelWidgets():
                        if isinstance(widget, QW.QMainWindow):
                            widget.statusBar().showMessage(message, msec)
                QC.QTimer.singleShot(
                    execenv.delay * 1000,
                    lambda: close_widgets_and_quit(screenshot=execenv.screenshot),
                )
            if exec_loop:
                QAPP_INSTANCE.exec()

    if enable_logs and Conf.main.faulthandler_enabled.get():
        faulthandler.disable()
    remove_empty_log_file(fh_log_fname)
    if enable_logs:
        logging.shutdown()
        remove_empty_log_file(tb_log_fname)


def close_dialog_and_quit(widget, screenshot=False):
    """Close QDialog and quit Qt event loop"""
    wname = widget.objectName()
    if screenshot and wname and widget.isVisible():  # pragma: no cover
        grab_save_window(widget, wname.lower())
    widget.done(QW.QDialog.Accepted)
    # QW.QApplication.instance().quit()


def exec_dialog(dlg):
    """Run QDialog Qt execution loop without blocking,
    depending on environment test mode"""
    if execenv.unattended:
        QC.QTimer.singleShot(
            execenv.delay * 1000,
            lambda: close_dialog_and_quit(dlg, screenshot=execenv.screenshot),
        )
    return dlg.exec()


def qt_wait(timeout, except_unattended=False):  # pragma: no cover
    """Freeze GUI during timeout (seconds) while processing Qt events"""
    if except_unattended and execenv.unattended:
        return
    start = time.time()
    while time.time() <= start + timeout:
        time.sleep(0.01)
        QW.QApplication.processEvents()


@contextmanager
def create_progress_bar(parent, label, max_):
    """Create modal progress bar"""
    prog = QW.QProgressDialog(label, _("Cancel"), 0, max_, parent, QC.Qt.SplashScreen)
    prog.setWindowModality(QC.Qt.WindowModal)
    prog.show()
    QW.QApplication.processEvents()
    try:
        yield prog
    finally:
        prog.close()


def qt_handle_error_message(widget, message):
    """Handles application (QWidget) error message"""
    traceback.print_exc()
    txt = str(message)
    msglines = txt.splitlines()
    if len(msglines) > 10:
        txt = os.linesep.join(msglines[:10] + ["..."])
    title = widget.window().objectName()
    QW.QMessageBox.critical(widget, title, _("Error:") + f"\n{txt}")


def qt_try_except(message=None):
    """Try...except Qt widget method decorator"""

    def qt_try_except_decorator(func):
        """Try...except Qt widget method decorator"""

        @functools.wraps(func)
        def method_wrapper(*args, **kwargs):
            """Decorator wrapper function"""
            self = args[0]  # extracting 'self' from method arguments
            #  If "self" is a BaseProcessor, then we need to get the panel instance
            panel = getattr(self, "panel", self)
            if message is not None:
                panel.SIG_STATUS_MESSAGE.emit(message)
                QW.QApplication.setOverrideCursor(QG.QCursor(QC.Qt.WaitCursor))
                panel.repaint()
            output = None
            try:
                output = func(*args, **kwargs)
            except Exception as msg:  # pylint: disable=broad-except
                qt_handle_error_message(panel.parent(), msg)
            finally:
                panel.SIG_STATUS_MESSAGE.emit("")
                QW.QApplication.restoreOverrideCursor()
            return output

        return method_wrapper

    return qt_try_except_decorator


@contextmanager
def qt_try_loadsave_file(widget: QW.QWidget, filename: str, operation: str):
    """Try and open file (operation: "load" or "save")"""
    if operation == "load":
        text = _("%s could not be opened:")
    elif operation == "save":
        text = _("%s could not be written:")
    else:
        raise ValueError("operation argument must be 'load' or 'save'")
    try:
        yield filename
    except Exception as msg:  # pylint: disable=broad-except
        traceback.print_exc()
        message = (text % osp.basename(filename)) + "\n" + str(msg)
        QW.QMessageBox.critical(widget, APP_NAME, message)
    finally:
        pass


SHOTPATH = osp.join(
    get_module_data_path("codraft"), os.pardir, "doc", "images", "shots"
)


def grab_save_window(widget: QW.QWidget, name: str) -> None:  # pragma: no cover
    """Grab window screenshot and save it"""
    widget.activateWindow()
    widget.raise_()
    QW.QApplication.processEvents()
    pixmap = widget.grab()
    suffix = ""
    if not name[-1].isdigit() and not name.startswith(("s_", "i_")):
        suffix = "_" + datetime.now().strftime("%Y-%m-%d-%H%M%S")
    pixmap.save(osp.join(SHOTPATH, f"{name}{suffix}.png"))


@contextmanager
def save_restore_stds():
    """Save/restore standard I/O before/after doing some things
    (e.g. calling Qt open/save dialogs)"""
    saved_in, saved_out, saved_err = sys.stdin, sys.stdout, sys.stderr
    sys.stdout = None
    try:
        yield
    finally:
        sys.stdin, sys.stdout, sys.stderr = saved_in, saved_out, saved_err


@contextmanager
def block_signals(widget: QW.QWidget, enable: bool):
    """Eventually block/unblock widget Qt signals before/after doing some things
    (enable: True if feature is enabled)"""
    if enable:
        widget.blockSignals(True)
    try:
        yield
    finally:
        if enable:
            widget.blockSignals(False)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT environmnent utilities
"""

import argparse
import enum
import os
import platform
import pprint
import sys


class VerbosityLevels(enum.Enum):
    """Print verbosity levels (for testing purpose)"""

    QUIET = "quiet"
    MINIMAL = "minimal"
    NORMAL = "normal"


class CodraFTExecEnv:
    """Object representing CodraFT test environment"""

    UNATTENDED_ARG = "unattended"
    VERBOSE_ARG = "verbose"
    SCREENSHOT_ARG = "screenshot"
    DELAY_ARG = "delay"
    UNATTENDED_ENV = "CODRAFT_UNATTENDED_TESTS"
    VERBOSE_ENV = "CODRAFT_VERBOSITY_LEVEL"
    SCREENSHOT_ENV = "CODRAFT_TAKE_SCREENSHOT"
    DELAY_ENV = "CODRAFT_DELAY_BEFORE_QUIT"

    def __init__(self):
        self.h5files = None
        self.h5browser_file = None
        self.demo_mode = False
        self.parse_args()

    def enable_demo_mode(self, delay: int):
        """Enable demo mode"""
        self.demo_mode = True
        self.unattended = True
        self.delay = delay

    @staticmethod
    def __get_mode(env):
        """Get mode value"""
        return os.environ.get(env) is not None

    @staticmethod
    def __set_mode(env, value):
        """Set mode value"""
        if env in os.environ:
            os.environ.pop(env)
        if value:
            os.environ[env] = "1"

    @property
    def unattended(self):
        """Get unattended value"""
        return self.__get_mode(self.UNATTENDED_ENV)

    @unattended.setter
    def unattended(self, value):
        """Set unattended value"""
        self.__set_mode(self.UNATTENDED_ENV, value)

    @property
    def screenshot(self):
        """Get screenshot value"""
        return self.__get_mode(self.SCREENSHOT_ENV)

    @screenshot.setter
    def screenshot(self, value):
        """Set screenshot value"""
        self.__set_mode(self.SCREENSHOT_ENV, value)
        if value:  # pragma: no cover
            self.unattended = value

    @property
    def verbose(self):
        """Get verbosity level"""
        return os.environ.get(self.VERBOSE_ENV, VerbosityLevels.NORMAL.value)

    @verbose.setter
    def verbose(self, value):
        """Set verbosity level"""
        os.environ[self.VERBOSE_ENV] = value

    @property
    def delay(self):
        """Delay (seconds) before quitting application in unattended mode"""
        try:
            return int(os.environ.get(self.DELAY_ENV))
        except (TypeError, ValueError):
            return 0

    @delay.setter
    def delay(self, value: int):
        """Set delay (seconds) before quitting application in unattended mode"""
        os.environ[self.DELAY_ENV] = str(value)

    def parse_args(self):
        """Parse command line arguments"""
        parser = argparse.ArgumentParser(description="Run CodraFT")
        parser.add_argument(
            "h5",
            nargs="?",
            type=str,
            help="HDF5 file names (separated by ';'), "
            "optionally with dataset name (separated by ',')",
        )
        parser.add_argument(
            "-b",
            "--h5browser",
            required=False,
            type=str,
            metavar="path",
            help="path to open with HDF5 browser",
        )
        parser.add_argument(
            "-v", "--version", action="store_true", help="show CodraFT version"
        )
        parser.add_argument(
            "--mode",
            choices=[self.UNATTENDED_ARG, self.SCREENSHOT_ARG],
            required=False,
            help="unattended: non-interactive test mode ; "
            "screenshot: unattended mode, with automatic screenshots",
        )
        parser.add_argument(
            "--" + self.DELAY_ARG,
            type=int,
            default=0,
            help="delay (seconds) before quitting application in unattended mode",
        )
        parser.add_argument(
            "--" + self.VERBOSE_ARG,
            choices=[lvl.value for lvl in VerbosityLevels],
            required=False,
            default=VerbosityLevels.NORMAL.value,
            help="verbosity level: for debugging/testing purpose",
        )
        args, _unknown = parser.parse_known_args()
        if args.h5:
            self.h5files = args.h5.split(";")
        if args.h5browser:
            self.h5browser_file = args.h5browser
        if args.version:
            version = os.environ["CODRAFT_VERSION"]
            print(f"CodraFT {version} on {platform.system()}")
            sys.exit()
        self.set_env_from_args(args)

    def set_env_from_args(self, args):
        """Set appropriate environment variables"""
        if args.mode is not None:
            self.unattended = args.mode == self.UNATTENDED_ARG
            self.screenshot = args.mode == self.SCREENSHOT_ARG
        if args.verbose is not None:
            self.verbose = args.verbose
        self.delay = args.delay

    def print(self, *objects, sep=" ", end="\n", file=sys.stdout, flush=False):
        """Print in file, depending on verbosity level"""
        # print(f"unattended={self.unattended} ; verbose={self.verbose} ; ")
        # print(f"screenshot={self.screenshot}; delay={self.delay}")
        if (self.verbose != VerbosityLevels.QUIET.value) and (
            self.verbose != VerbosityLevels.MINIMAL.value or file == sys.stderr
        ):
            print(*objects, sep=sep, end=end, file=file, flush=flush)
        # TODO: [P4] Eventually add logging here

    def pprint(
        self,
        obj,
        stream=None,
        indent=1,
        width=80,
        depth=None,
        compact=False,
        sort_dicts=True,
    ):
        """Pretty-print in stream, depending on verbosity level"""
        if (self.verbose != VerbosityLevels.QUIET.value) and (
            self.verbose != VerbosityLevels.MINIMAL.value or stream == sys.stderr
        ):
            pprint.pprint(
                obj,
                stream=stream,
                indent=indent,
                width=width,
                depth=depth,
                compact=compact,
                sort_dicts=sort_dicts,
            )


execenv = CodraFTExecEnv()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Annotations unit test:

  - Create an image with annotations
  - Open dialog (equivalent to click on button "Annotations")
  - Accept dialog without modifying shapes
  - Check if image annotations are still the same
"""

from qtpy import QtWidgets as QW

from codraft.core.model.base import ANN_KEY
from codraft.env import execenv
from codraft.tests import codraft_app_context
from codraft.tests import data as test_data

SHOW = True  # Show test in GUI-based test launcher


def test():
    """Run image tools test scenario"""
    execenv.unattended = True
    with codraft_app_context() as win:
        panel = win.imagepanel
        ima = test_data.create_image_with_annotations()
        panel.add_object(ima)
        panel.open_separate_view().done(QW.QDialog.DialogCode.Accepted)
        orig_metadata = ima.metadata.copy()
        panel.open_separate_view().done(QW.QDialog.DialogCode.Accepted)
        execenv.print("Check [geometric shapes] <--> [plot items] conversion:")
        execenv.print(f"  Comparing {ANN_KEY}: ", end="")
        # open("before.json", mode="wb").write(orig_metadata[ANN_KEY].encode())
        # open("after.json", mode="wb").write(ima.metadata[ANN_KEY].encode())
        assert orig_metadata[ANN_KEY] == ima.metadata[ANN_KEY]
        execenv.print("OK")


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT Datasets
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

import abc
import enum
import json
import sys

import guidata.dataset.dataitems as gdi
import guidata.dataset.datatypes as gdt
import numpy as np
from guidata.jsonio import JSONHandler, JSONReader, JSONWriter
from guiqwt.annotations import (
    AnnotatedCircle,
    AnnotatedEllipse,
    AnnotatedPoint,
    AnnotatedShape,
)
from guiqwt.builder import make
from guiqwt.io import load_items, save_items
from guiqwt.styles import AnnotationParam

from codraft.config import Conf, _
from codraft.utils.misc import is_integer_dtype

ROI_KEY = "_roi_"
ANN_KEY = "_ann_"


class MetadataItem(gdt.DataItem):
    """
    Construct a data item representing a metadata dictionary
        * label [string]: name
        * default [dict]: default value (optional)
        * help [string]: text shown in tooltip (optional)
        * check [bool]: if False, value is not checked (optional, default=True)
    """

    # pylint: disable=redefined-builtin,abstract-method
    def __init__(self, label, default=None, help="", check=True):
        gdt.DataItem.__init__(self, label, default=default, help=help, check=check)
        self.set_prop("display", callback=self.__dictedit)
        self.set_prop("display", icon="dictedit.png")

    @staticmethod
    # pylint: disable=unused-argument
    def __dictedit(instance, item, value, parent):
        """Open a dictionary editor"""
        # pylint: disable=import-outside-toplevel
        from guidata.widgets.collectionseditor import CollectionsEditor

        editor = CollectionsEditor(parent)
        value_was_none = value is None
        if value_was_none:
            value = {}
        editor.setup(value)
        if editor.exec():
            return editor.get_value()
        if value_was_none:
            return None
        return value

    def serialize(self, instance, writer):
        """Serialize this item"""
        value = self.get_value(instance)
        writer.write_dict(value)

    def get_value_from_reader(self, reader):
        """Reads value from the reader object, inside the try...except
        statement defined in the base item `deserialize` method"""
        return reader.read_dict()


@enum.unique
class Choices(enum.Enum):
    """Object associating an enum to guidata.dataset.dataitems.ChoiceItem choices"""

    # Reimplement enum.Enum method as suggested by Python documentation:
    # https://docs.python.org/3/library/enum.html#using-automatic-values
    # Here, it is only needed for ImageDatatypes (see core/model/image.py).
    # pylint: disable=unused-argument,no-self-argument,no-member
    def _generate_next_value_(name, start, count, last_values):
        return name.lower()

    @classmethod
    def get_choices(cls):
        """Return tuple of (key, value) choices to be used as parameter of
        guidata.dataset.dataitems.ChoiceItem"""
        return tuple((member, member.value) for member in cls)


class BaseProcParam(gdt.DataSet):
    """Base class for processing parameters"""

    def __init__(self, title=None, comment=None, icon=""):
        super().__init__(title, comment, icon)
        self.set_global_prop("data", min=None, max=None)

    def apply_integer_range(self, vmin, vmax):  # pylint: disable=unused-argument
        """Do something in case of integer min-max range"""

    def apply_float_range(self, vmin, vmax):  # pylint: disable=unused-argument
        """Do something in case of float min-max range"""

    def set_from_datatype(self, dtype):
        """Set min/max range from NumPy datatype"""
        if is_integer_dtype(dtype):
            info = np.iinfo(dtype)
            self.apply_integer_range(info.min, info.max)
        else:
            info = np.finfo(dtype)
            self.apply_float_range(info.min, info.max)
        self.set_global_prop("data", min=info.min, max=info.max)


class BaseRandomParam(BaseProcParam):
    """Random signal/image parameters"""

    seed = gdi.IntItem(_("Seed"), default=1)


class UniformRandomParam(BaseRandomParam):
    """Uniform-law random signal/image parameters"""

    def apply_integer_range(self, vmin, vmax):
        """Do something in case of integer min-max range"""
        self.vmin, self.vmax = vmin, vmax

    vmin = gdi.FloatItem(
        "V<sub>min</sub>", default=-0.5, help=_("Uniform distribution lower bound")
    )
    vmax = gdi.FloatItem(
        "V<sub>max</sub>", default=0.5, help=_("Uniform distribution higher bound")
    ).set_pos(col=1)


class NormalRandomParam(BaseRandomParam):
    """Normal-law random signal/image parameters"""

    DEFAULT_RELATIVE_MU = 0.1
    DEFAULT_RELATIVE_SIGMA = 0.02

    def apply_integer_range(self, vmin, vmax):
        """Do something in case of integer min-max range"""
        delta = vmax - vmin
        self.mu = int(self.DEFAULT_RELATIVE_MU * delta + vmin)
        self.sigma = int(self.DEFAULT_RELATIVE_SIGMA * delta)

    mu = gdi.FloatItem(
        "μ", default=DEFAULT_RELATIVE_MU, help=_("Normal distribution mean")
    )
    sigma = gdi.FloatItem(
        "σ",
        default=DEFAULT_RELATIVE_SIGMA,
        help=_("Normal distribution standard deviation"),
    ).set_pos(col=1)


@enum.unique
class ShapeTypes(enum.Enum):
    """Shape types for image metadata"""

    # Reimplement enum.Enum method as suggested by Python documentation:
    # https://docs.python.org/3/library/enum.html#using-automatic-values
    # pylint: disable=unused-argument,no-self-argument,no-member
    def _generate_next_value_(name, start, count, last_values):
        return f"_{name.lower()[:3]}_"

    RECTANGLE = enum.auto()
    CIRCLE = enum.auto()
    ELLIPSE = enum.auto()
    SEGMENT = enum.auto()
    MARKER = enum.auto()
    POINT = enum.auto()


def config_annotated_shape(item: AnnotatedShape, fmt: str, lbl: bool, cmp: bool = None):
    """Configurate annotated shape"""
    param = item.annotationparam
    param.format = fmt
    param.show_label = lbl
    if cmp is not None:
        param.show_computations = cmp
    param.update_annotation(item)


def set_plot_item_editable(item, state):
    """Set plot item editable state"""
    item.set_movable(state)
    item.set_resizable(state)
    item.set_rotatable(state)
    item.set_readonly(not state)


# TODO: [P0] Replace 'array' by 'datalist', a list of NumPy arrays
# With this new data model, the old 'array' attribute row (each row is a result) is
# replaced by an element of the new 'datalist' attribute. So, when this change is done,
# each 'datalist' element is a result. This means that each result no longer has to be
# an array with the same number of columns: in other words, each result may be an
# arbitrary NumPy array, with an arbitrary shape. This is the opportunity to introduce
# a new ShapeTypes type (e.g. FREEFORM) represented by an AnnotatedPolygon (new class
# to be written using AnnotatedRectangle as a model). This also has been made possible
# due to a recent change in CodraFT HDF5 (de)serialization which now accepts nested
# lists or dictionnaries.
class ResultShape:
    """Object representing a geometrical shape serializable in signal/image metadata.

    Result `array` is a NumPy 2-D array: each row is a result, optionnally associated
    to a ROI (first column value).

    ROI index is starting at 0 (or is simply 0 if there is no ROI).

    :param ShapeTypes shapetype: shape type
    :param np.ndarray array: shape coordinates (multiple shapes: one shape per row),
    first column is ROI index (0 if there is no ROI)
    :param str label: shape label
    """

    def __init__(self, shapetype: ShapeTypes, array: np.ndarray, label: str = ""):
        assert isinstance(label, str)
        assert isinstance(shapetype, ShapeTypes)
        self.label = self.show_label = label
        self.shapetype = shapetype
        if isinstance(array, (list, tuple)):
            if isinstance(array[0], (list, tuple)):
                array = np.array(array)
            else:
                array = np.array([array])
        assert isinstance(array, np.ndarray)
        self.array = array
        if label.endswith("s"):
            self.show_label = label[:-1]
        self.check_array()

    @classmethod
    def label_shapetype_from_key(cls, key: str):
        """Return metadata shape label and shapetype from metadata key"""
        for member in ShapeTypes:
            if key.startswith(member.value):
                label = key[len(member.value) :]
                return label, member
        raise ValueError(f"Invalid metadata key `{key}`")

    @classmethod
    def from_metadata_entry(cls, key, value):
        """Create metadata shape object from (key, value) metadata entry"""
        if isinstance(key, str) and isinstance(value, np.ndarray):
            try:
                label, shapetype = cls.label_shapetype_from_key(key)
                return cls(shapetype, value, label)
            except ValueError:
                pass
        return None

    @classmethod
    def match(cls, key, value):
        """Return True if metadata dict entry (key, value) is a metadata result"""
        return cls.from_metadata_entry(key, value) is not None

    @property
    def key(self):
        """Return metadata key associated to result"""
        return self.shapetype.value + self.label

    @property
    def xlabels(self):
        """Return labels for result array columns"""
        if self.shapetype in (ShapeTypes.MARKER, ShapeTypes.POINT):
            labels = "ROI", "x", "y"
        elif self.shapetype in (
            ShapeTypes.RECTANGLE,
            ShapeTypes.CIRCLE,
            ShapeTypes.SEGMENT,
        ):
            labels = "ROI", "x0", "y0", "x1", "y1"
        elif self.shapetype is ShapeTypes.ELLIPSE:
            labels = "ROI", "x0", "y0", "x1", "y1", "x2", "y2", "x3", "y3"
        else:
            raise NotImplementedError(f"Unsupported shapetype {self.shapetype}")
        return labels[-self.array.shape[1] :]

    def add_to(self, obj):
        """Add metadata shape to object (signal/image)"""
        obj.metadata[self.key] = self.array
        if self.shapetype in (
            ShapeTypes.SEGMENT,
            ShapeTypes.CIRCLE,
            ShapeTypes.ELLIPSE,
        ):
            #  Automatically adds segment norm / circle diameter to object metadata
            colnb = 2
            if self.shapetype is ShapeTypes.ELLIPSE:
                colnb += 1
            arr = self.array
            results = np.zeros((arr.shape[0], colnb), dtype=arr.dtype)
            results[:, 0] = arr[:, 0]  # ROI indexes
            dx1, dy1 = arr[:, 3] - arr[:, 1], arr[:, 4] - arr[:, 2]
            results[:, 1] = np.linalg.norm(np.vstack([dx1, dy1]).T, axis=1)
            if self.shapetype is ShapeTypes.ELLIPSE:
                dx2, dy2 = arr[:, 7] - arr[:, 5], arr[:, 8] - arr[:, 6]
                results[:, 2] = np.linalg.norm(np.vstack([dx2, dy2]).T, axis=1)
            label = self.label
            if self.shapetype is ShapeTypes.CIRCLE:
                label += "Diameter"
            if self.shapetype is ShapeTypes.ELLIPSE:
                label += "Diameters"
            obj.metadata[label] = results

    def merge_with(self, obj, other_obj=None):
        """Merge object resultshape with another's: obj <-- other_obj
        or simply merge this resultshape with obj if other_obj is None"""
        if other_obj is None:
            other_obj = obj
        other_value = other_obj.metadata.get(self.key)
        if other_value is not None:
            other = ResultShape.from_metadata_entry(self.key, other_value)
            other_array = np.array(other.array, copy=True)
            if other_array.shape[1] > self.data_colnb:  # Column 0 is the ROI index
                other_array[:, 0] += self.array[-1, 0] + 1  # Adding ROI index offset
            self.array = np.vstack([self.array, other_array])
        self.add_to(obj)

    @property
    def data_colnb(self):
        """Return raw data results column number"""
        return {
            ShapeTypes.MARKER: 2,
            ShapeTypes.POINT: 2,
            ShapeTypes.RECTANGLE: 4,
            ShapeTypes.CIRCLE: 4,
            ShapeTypes.SEGMENT: 4,
            ShapeTypes.ELLIPSE: 8,
        }[self.shapetype]

    @property
    def data(self):
        """Return raw data (array without ROI informations)"""
        return self.array[:, -self.data_colnb :]

    def check_array(self):
        """Check if array is valid"""
        assert len(self.array.shape) == 2
        assert self.array.shape[1] == self.data_colnb + 1

    def iterate_plot_items(self, fmt: str, lbl: bool):
        """Iterate over metadata shape plot items

        :param str fmt: numeric format (e.g. "%.3f")
        :param bool lbl: if True, show shape labels
        """
        for args in self.data:
            yield self.create_plot_item(args, fmt, lbl)

    def create_plot_item(self, args: np.ndarray, fmt: str, lbl: bool):
        """Make plot item"""
        if self.shapetype is ShapeTypes.MARKER:
            item = self.make_marker_item(args, fmt)
        elif self.shapetype is ShapeTypes.POINT:
            item = AnnotatedPoint(*args)
            sparam = item.shape.shapeparam
            sparam.symbol.marker = "Ellipse"
            sparam.symbol.size = 6
            sparam.sel_symbol.marker = "Ellipse"
            sparam.sel_symbol.size = 6
            sparam.update_shape(item.shape)
            param = item.annotationparam
            param.title = self.show_label
            param.update_annotation(item)
        elif self.shapetype is ShapeTypes.RECTANGLE:
            x0, y0, x1, y1 = args
            item = make.annotated_rectangle(x0, y0, x1, y1, title=self.show_label)
        elif self.shapetype is ShapeTypes.CIRCLE:
            x0, y0, x1, y1 = args
            param = AnnotationParam(_("Annotation"), icon="annotation.png")
            param.title = self.show_label
            item = AnnotatedCircle(x0, y0, x1, y1, param)
            item.set_style("plot", "shape/drag")
        elif self.shapetype is ShapeTypes.SEGMENT:
            x0, y0, x1, y1 = args
            item = make.annotated_segment(x0, y0, x1, y1, title=self.show_label)
        elif self.shapetype is ShapeTypes.ELLIPSE:
            x0, y0, x1, y1, x2, y2, x3, y3 = args
            param = AnnotationParam(_("Annotation"), icon="annotation.png")
            param.title = self.show_label
            item = AnnotatedEllipse(annotationparam=param)
            item.shape.switch_to_ellipse()
            item.set_xdiameter(x0, y0, x1, y1)
            item.set_ydiameter(x2, y2, x3, y3)
            item.set_style("plot", "shape/drag")
        else:
            print(f"Warning: unsupported item {self.shapetype}", file=sys.stderr)
            return None
        if isinstance(item, AnnotatedShape):
            config_annotated_shape(item, fmt, lbl)
        set_plot_item_editable(item, False)
        return item

    def make_marker_item(self, args, fmt):
        """Make marker item"""
        x0, y0 = args
        if np.isnan(x0):
            mstyle = "-"

            def label(x, y):  # pylint: disable=unused-argument
                return (self.show_label + ": " + fmt) % y

        elif np.isnan(y0):
            mstyle = "|"

            def label(x, y):  # pylint: disable=unused-argument
                return (self.show_label + ": " + fmt) % x

        else:
            mstyle = "+"
            txt = self.show_label + ": (" + fmt + ", " + fmt + ")"

            def label(x, y):
                return txt % (x, y)

        return make.marker(
            position=(x0, y0),
            markerstyle=mstyle,
            label_cb=label,
            linestyle="DashLine",
            color="yellow",
        )


def make_roi_item(func, coords: list, title: str, fmt: str, lbl: bool, editable: bool):
    """Make ROI item shape"""
    item = func(*coords, title)
    if not editable:
        if isinstance(item, AnnotatedShape):
            config_annotated_shape(item, fmt, lbl, cmp=editable)
            item.set_style("plot", "shape/mask")
        item.set_movable(False)
        item.set_resizable(False)
        item.set_readonly(True)
    return item


class ObjectItfMeta(abc.ABCMeta, gdt.DataSetMeta):
    """Mixed metaclass to avoid conflicts"""


class ObjectItf(metaclass=ObjectItfMeta):
    """Object (signal/image) interface"""

    metadata = {}  # This is overriden in children classes with a gdi.DictItem instance

    # Metadata dictionary keys for special properties:
    METADATA_FMT = "__format"
    METADATA_LBL = "__showlabel"

    DEFAULT_FMT = "s"  # This is overriden in children classes
    CONF_FMT = Conf.view.sig_format  # This is overriden in children classes
    VALID_DTYPES = ()

    @property
    @abc.abstractmethod
    def data(self):
        """Data"""

    def check_data(self):
        """Check if data is valid, raise an exception if that's not the case"""
        if self.data is not None:
            if self.data.dtype not in self.VALID_DTYPES:
                raise TypeError(f"Unsupported data type: {self.data.dtype}")

    def iterate_roi_indexes(self):
        """Iterate over object ROI indexes ([0] if there is no ROI)"""
        if self.roi is None:
            yield 0
        else:
            yield from range(len(self.roi))

    @abc.abstractmethod
    def get_data(self, roi_index: int = None) -> np.ndarray:
        """
        Return original data (if ROI is not defined or `roi_index` is None),
        or ROI data (if both ROI and `roi_index` are defined).
        """

    @abc.abstractmethod
    def copy_data_from(self, other, dtype=None):
        """Copy data from other dataset instance"""

    @abc.abstractmethod
    def set_data_type(self, dtype):
        """Change data type"""

    @abc.abstractmethod
    def make_item(self, update_from=None):
        """Make plot item from data"""

    @abc.abstractmethod
    def update_item(self, item, ref_item=None):
        """Update plot item from data"""

    @abc.abstractmethod
    def roi_coords_to_indexes(self, coords: list) -> np.ndarray:
        """Convert ROI coordinates to indexes"""

    @abc.abstractmethod
    def get_roi_param(self, title, *defaults):
        """Return ROI parameters dataset"""

    def roidata_to_params(self, roidata: np.ndarray) -> gdt.DataSetGroup:
        """Convert ROI array data to ROI dataset group"""
        roi_params = []
        for index, parameters in enumerate(roidata):
            roi_param = self.get_roi_param(f"ROI{index:02d}", *parameters)
            roi_params.append(roi_param)
        group = gdt.DataSetGroup(roi_params, title=_("Regions of interest"))
        return group

    @staticmethod
    @abc.abstractmethod
    def params_to_roidata(params: gdt.DataSetGroup) -> np.ndarray:
        """Convert ROI dataset group to ROI array data"""

    @property
    def roi(self) -> np.ndarray:
        """Return object regions of interest array (one ROI per line)"""
        return self.metadata.get(ROI_KEY)

    @roi.setter
    def roi(self, roidata: np.ndarray):
        """Set object regions of interest array, using a list or ROI dataset params"""
        if roidata is None:
            if ROI_KEY in self.metadata:
                self.metadata.pop(ROI_KEY)
        else:
            self.metadata[ROI_KEY] = np.array(roidata, int)

    def add_resultshape(
        self, label: str, shapetype: ShapeTypes, array: np.ndarray
    ) -> ResultShape:
        """Add geometric shape as metadata entry, and return ResultShape instance"""
        mshape = ResultShape(shapetype, array, label)
        mshape.add_to(self)
        return mshape

    def update_resultshapes_from(self, other):
        """Update geometric shape from another object (merge metadata)"""
        for key, value in self.metadata.items():
            if ResultShape.match(key, value):
                mshape = ResultShape.from_metadata_entry(key, value)
                mshape.merge_with(self, other)

    @abc.abstractmethod
    def iterate_roi_items(self, fmt: str, lbl: bool, editable: bool = True):
        """Make plot item representing a Region of Interest"""

    def __set_annotations(self, annotations: str):
        """Set object annotations (JSON string describing annotation plot items)"""
        self.metadata[ANN_KEY] = annotations

    def __get_annotations(self) -> str:
        """Get object annotations (JSON string describing annotation plot items)"""
        return self.metadata.get(ANN_KEY, "")

    annotations = property(__get_annotations, __set_annotations)

    def set_annotations_from_items(self, items: list):
        """Set object annotations (annotation plot items)"""
        writer = JSONWriter(None)
        save_items(writer, items)
        self.annotations = writer.get_json(indent=4)

    def iterate_shape_items(self, editable: bool = False):
        """Iterate over computing items encoded in metadata (if any)"""
        setdef = self.metadata.setdefault
        fmt = setdef(self.METADATA_FMT, "%" + self.CONF_FMT.get(self.DEFAULT_FMT))
        lbl = setdef(self.METADATA_LBL, Conf.view.show_label.get(False))
        for key, value in self.metadata.items():
            if key == ROI_KEY:
                yield from self.iterate_roi_items(fmt=fmt, lbl=lbl, editable=False)
            elif ResultShape.match(key, value):
                mshape = ResultShape.from_metadata_entry(key, value)
                yield from mshape.iterate_plot_items(fmt, lbl)
        if self.annotations:
            try:
                for item in load_items(JSONReader(self.annotations)):
                    set_plot_item_editable(item, editable)
                    if isinstance(item, AnnotatedShape):
                        config_annotated_shape(item, fmt, lbl)
                    yield item
            except json.decoder.JSONDecodeError:
                pass

    def remove_resultshapes(self):
        """Remove metadata shapes and ROIs"""
        for key, value in list(self.metadata.items()):
            resultshape = ResultShape.from_metadata_entry(key, value)
            if resultshape is not None or key == ROI_KEY:
                # Metadata entry is a metadata shape or a ROI
                self.metadata.pop(key)

    def export_metadata_to_file(self, filename):
        """Export object metadata to file (JSON)"""
        handler = JSONHandler(filename)
        handler.set_json_dict(self.metadata)
        handler.save()

    def import_metadata_from_file(self, filename):
        """Import object metadata from file (JSON)"""
        handler = JSONHandler(filename)
        handler.load()
        self.metadata = handler.get_json_dict()

    def metadata_to_html(self):
        """Convert metadata to human-readable string"""
        textlines = []
        for key, value in self.metadata.items():
            if len(textlines) > 5:
                textlines.append("[...]")
                break
            if not key.startswith("_"):
                vlines = str(value).splitlines()
                if vlines:
                    text = f"<b>{key}:</b> {vlines[0]}"
                    if len(vlines) > 1:
                        text += " [...]"
                    textlines.append(text)
        if textlines:
            ptit = _("Object metadata")
            psub = _("(click on Metadata button for more details)")
            prefix = f"<i><u>{ptit}:</u> {psub}</i><br>"
            return f"<p style='white-space:pre'>{prefix}{'<br>'.join(textlines)}</p>"
        return ""



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT I/O conversion functions
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from typing import List

import numpy as np


def data_to_xy(data: np.ndarray) -> List[np.ndarray]:
    """Convert 2-D array into a list of 1-D array data (x, y, dx, dy).
    This is useful for importing data and creating a CodraFT signal with it."""
    rows, cols = data.shape
    for colnb in (2, 3, 4):
        if cols == colnb and rows > colnb:
            data = data.T
            break
    if len(data) == 1:
        data = data.T
    if len(data) not in (2, 3, 4):
        raise ValueError(f"Invalid data: len(data)={len(data)} (expected 2, 3 or 4)")
    x, y = data[:2]
    dx, dy = None, None
    if len(data) == 3:
        dy = data[2]
    if len(data) == 4:
        dx, dy = data[2:]
    return x, y, dx, dy



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Application embedded test 1

CodraFT main window is destroyed when closing application.
It is rebuilt from scratch when reopening application.
"""

from guidata.qthelpers import get_std_icon
from guidata.widgets.codeeditor import CodeEditor
from qtpy import QtCore as QC
from qtpy import QtWidgets as QW

from codraft.config import _
from codraft.core.gui.main import CodraFTMainWindow
from codraft.tests import data as test_data
from codraft.utils.qthelpers import qt_app_context

SHOW = True  # Show test in GUI-based test launcher


class HostWidget(QW.QWidget):
    """Host widget: menu with action buttons, log viewer"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.button_layout = QW.QVBoxLayout()
        self.logwidget = CodeEditor(self)
        self.logwidget.setMinimumWidth(500)
        grid_layout = QW.QGridLayout()
        grid_layout.addLayout(self.button_layout, 0, 0)
        grid_layout.addWidget(self.logwidget, 0, 1)
        self.setLayout(grid_layout)

    def log(self, message):
        """Log message"""
        self.logwidget.appendPlainText(message)

    def add_spacing(self, spacing: int) -> None:
        """Add spacing to button box"""
        self.button_layout.addSpacing(spacing)

    def add_label(self, text: str) -> None:
        """Add label to button box"""
        self.button_layout.addWidget(QW.QLabel(text))

    def add_widget(self, obj: QW.QWidget, spacing_before: int = 0) -> None:
        """Add widget (QWidget) to button box"""
        if spacing_before > 0:
            self.add_spacing(spacing_before)
        self.button_layout.addWidget(obj)

    def add_button(self, title, slot, spacing_before=0, icon=None):
        """Add button"""
        btn = QW.QPushButton(title)
        if icon is not None:
            btn.setIcon(get_std_icon(icon))
        btn.clicked.connect(slot)
        self.add_widget(btn, spacing_before=spacing_before)

    def add_stretch(self):
        """Add stretch to button box"""
        self.button_layout.addStretch()


class BaseHostWindow(QW.QMainWindow):
    """Main window"""

    SIG_TITLES = ("Oscilloscope", "Digitizer", "Radiometer", "Voltmeter", "Sensor")
    IMA_TITLES = (
        "Camera",
        "Streak Camera",
        "Image Scanner",
        "Laser Beam Profiler",
        "Gated Imaging Camera",
    )

    def __init__(self):
        super().__init__()
        self.setWindowTitle(_("Host application"))
        self.setWindowIcon(get_std_icon("ComputerIcon"))
        self.codraft = None
        self.host = HostWidget(self)
        self.setCentralWidget(self.host)
        self.setup_window()
        self.host.add_stretch()
        self.index_sigtitle = -1
        self.index_imatitle = -1

    @property
    def sigtitle(self):
        """Return current signal title index"""
        self.index_sigtitle = idx = (self.index_sigtitle + 1) % len(self.SIG_TITLES)
        return self.SIG_TITLES[idx]

    @property
    def imatitle(self):
        """Return current image title index"""
        self.index_imatitle = idx = (self.index_imatitle + 1) % len(self.IMA_TITLES)
        return self.IMA_TITLES[idx]

    def setup_window(self):
        """Setup window"""
        self.host.add_label(_("This the host application, which embeds CodraFT."))
        add_btn = self.host.add_button
        add_btn(_("Open CodraFT"), self.open_codraft, 10, "DialogApplyButton")
        add_btn(_("Import signal from CodraFT"), self.import_signal, 10, "ArrowLeft")
        add_btn(_("Import image from CodraFT"), self.import_image, 0, "ArrowLeft")
        add_btn(_("Add signal objects"), self.add_signals, 10, "CommandLink")
        add_btn(_("Add image objects"), self.add_images, 0, "CommandLink")
        add_btn(_("Remove all objects"), self.remove_all, 5, "MessageBoxWarning")
        add_btn(_("Close CodraFT"), self.close_codraft, 10, "DialogCloseButton")

    def open_codraft(self):
        """Open CodraFT test"""
        raise NotImplementedError

    def close_codraft(self):
        """Close CodraFT window"""
        if self.codraft is not None:
            self.host.log("=> Closed CodraFT")
            self.codraft.close()
            self.codraft = None

    def import_object(self, panel, title):
        """Import object from CodraFT"""
        self.host.log(f"get_object_dialog ({title}):")
        obj = panel.get_object_dialog(self.host, title)
        if obj is not None:
            self.host.log(f"  -> {obj.title}:")
            self.host.log(str(obj))
        else:
            self.host.log("  -> canceled")

    def import_signal(self):
        """Import signal from CodraFT"""
        if self.codraft is not None:
            self.import_object(self.codraft.signalpanel, self.sender().text())

    def import_image(self):
        """Import image from CodraFT"""
        if self.codraft is not None:
            self.import_object(self.codraft.imagepanel, self.sender().text())

    def add_signals(self):
        """Add signals to CodraFT"""
        if self.codraft is not None:
            for func in (test_data.create_test_signal1, test_data.create_test_signal2):
                obj = func(title=self.sigtitle)
                self.codraft.signalpanel.add_object(obj)
                self.host.log(f"Added signal: {obj.title}")

    def add_images(self):
        """Add images to CodraFT"""
        if self.codraft is not None:
            size = 2000
            for func in (
                test_data.create_test_image1,
                test_data.create_test_image2,
                test_data.create_test_image3,
            ):
                obj = func(size, title=self.imatitle)
                self.codraft.imagepanel.add_object(obj)
                self.host.log(f"Added image: {obj.title}")

    def remove_all(self):
        """Remove all objects from CodraFT"""
        if self.codraft is not None:
            for panel in self.codraft.panels:
                objn = len(panel.objlist)
                panel.remove_all_objects()
                self.host.log(f"Removed {objn} objects from {panel.PANEL_STR}")


class HostWindow(BaseHostWindow):
    """Test main view"""

    def open_codraft(self):
        """Open CodraFT test"""
        if self.codraft is None:
            self.codraft = CodraFTMainWindow(console=False)
            self.codraft.setAttribute(QC.Qt.WA_DeleteOnClose, True)
            self.codraft.show()
            self.host.log("✨Initialized CodraFT window")
        else:
            try:
                self.codraft.show()
                self.codraft.raise_()
                self.host.log("=> Shown CodraFT window")
            except RuntimeError:
                self.codraft = None
                self.open_codraft()


def test_embedded_feature(klass):
    """Testing embedded feature"""
    with qt_app_context(exec_loop=True, enable_logs=False):
        window = klass()
        window.resize(800, 800)
        window.show()


if __name__ == "__main__":
    test_embedded_feature(HostWindow)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Statistics test

Testing the following:
  - Create a signal
  - Compute statistics on signal and show results
  - Create an image
  - Compute statistics on image and show results
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

import numpy as np

from codraft.config import _
from codraft.core.model import image as imod
from codraft.core.model import signal as smod
from codraft.tests import codraft_app_context, take_plotwidget_screenshot

SHOW = True  # Show test in GUI-based test launcher


def test():
    """Run statistics unit test scenario"""
    with codraft_app_context() as win:
        # === Signal statistics test ===
        panel = win.signalpanel
        snew = smod.new_signal_param(_("Gaussian"), stype=smod.SignalTypes.GAUSS)
        addparam = smod.GaussLorentzVoigtParam()
        sig = smod.create_signal_from_param(snew, addparam=addparam, edit=False)
        panel.add_object(sig)
        panel.processor.compute_stats()
        sig.roi = np.array([[len(sig.x) // 2, len(sig.x) - 1]], int)
        take_plotwidget_screenshot(panel, "stat_test")
        panel.processor.compute_stats()
        # === Image statistics test ===
        panel = win.imagepanel
        inew = imod.new_image_param(_("Raw data (2D-Gaussian)"), imod.ImageTypes.GAUSS)
        addparam = imod.Gauss2DParam()
        ima = imod.create_image_from_param(inew, addparam=addparam, edit=False)
        dy, dx = ima.size
        ima.roi = np.array(
            [
                [dx // 2, 0, dx, dy],
                [0, 0, dx // 3, dy // 3],
                [dx // 2, dy // 2, dx, dy],
            ],
            int,
        )
        panel.add_object(ima)
        take_plotwidget_screenshot(panel, "stat_test")
        panel.processor.compute_stats()


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT Common tools for exogenous HDF5 format support
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

import abc
import os.path as osp
from typing import Callable, Dict

import h5py
import numpy as np

from codraft.config import Conf
from codraft.core.io.conv import data_to_xy
from codraft.utils.misc import to_string


class BaseNode(metaclass=abc.ABCMeta):
    """Object representing a HDF5 node"""

    IS_ARRAY = False

    def __init__(self, h5file, dname):
        self.h5file = h5file
        self.dset = h5file[dname]
        self.metadata = {}
        self.__obj = None
        self.children = []
        self.uint32_wng = False

    @property
    def id(self):
        """Return node id"""
        return self.dset.name

    @property
    def name(self):
        """Return node name, constructed from dataset name"""
        return to_string(self.dset.name).split("/")[-1]

    @property
    def data(self):
        """Data associated to node, if available"""
        return None

    @property
    def icon_name(self):
        """Icon name associated to node"""

    @property
    def shape_str(self):
        """Return string representation of node shape, if any"""
        return ""

    @property
    def dtype_str(self):
        """Return string representation of node data type, if any"""
        return ""

    @property
    def text(self):
        """Return node textual representation"""

    @property
    def description(self):
        """Return node description"""
        return ""

    @classmethod
    def match(cls, dset):
        """Return True if h5 dataset match node pattern"""

    def create_object(self):  # pylint: disable=no-self-use
        """Create native object, if supported"""
        return None

    def get_object(self):
        """Return native object, if supported"""
        if self.__obj is None:
            obj = self.create_object()  # pylint: disable=assignment-from-none
            if obj is not None:
                self.__process_metadata(obj)
            self.__obj = obj
        return self.__obj

    def __process_metadata(self, obj):
        """Process metadata from dataset to obj"""
        obj.metadata = {}
        obj.metadata["HDF5Path"] = self.h5file.filename
        obj.metadata["HDF5Dataset"] = self.id
        for key, value in self.dset.attrs.items():
            if isinstance(value, bytes):
                value = to_string(value)
            obj.metadata[key] = value
        obj.metadata.update(self.metadata)

    @property
    def object_title(self):
        """Return signal/image object title"""
        if Conf.io.h5_fullpath_in_title.get(False):
            title = self.id
        else:
            title = self.name
        if Conf.io.h5_fname_in_title.get(True):
            title += f" ({osp.basename(self.h5file.filename)})"
        return title

    def set_signal_data(self, obj):
        """Set signal data (handles various issues)"""
        data = self.data
        if data.dtype not in (float, np.complex128):
            data = np.array(data, dtype=float)
        if len(data.shape) == 1:
            obj.set_xydata(np.arange(data.size), data)
        else:
            x, y, dx, dy = data_to_xy(data)
            obj.set_xydata(x, y, dx, dy)

    def set_image_data(self, obj):
        """Set image data (handles various issues)"""
        data = self.data
        if data.dtype == np.uint32:
            self.uint32_wng = data.max() > np.iinfo(np.int32).max
            clipped_data = data.clip(0, np.iinfo(np.int32).max)
            data = np.array(clipped_data, dtype=np.int32)
        obj.data = data


class H5Importer:
    """CodraFT HDF5 importer class"""

    def __init__(self, filename):
        self.h5file = h5py.File(filename)
        self.__nodes = {}
        self.root = RootNode(self.h5file)
        self.__nodes[self.root.id] = self.root.dset
        self.root.collect_children(self.__nodes)
        NODE_FACTORY.run_post_triggers(self)

    @property
    def nodes(self):
        """Return all nodes"""
        return self.__nodes.values()

    def get(self, node_id: str):
        """Return node associated to id"""
        return self.__nodes[node_id]

    def get_relative(self, node: BaseNode, relpath: str, ancestor: int = 0):
        """Return node using relative path to another node"""
        path = "/" + (
            "/".join(node.id.split("/")[:-ancestor]) + "/" + relpath.strip("/")
        ).strip("/")
        return self.__nodes[path]

    def close(self):
        """Close HDF5 file"""
        self.__nodes = {}
        self.h5file.close()


class NodeFactory:
    """Factory for node classes"""

    def __init__(self):
        self.__ignored_datasets = []
        self.__generic_classes = []
        self.__thirdparty_classes = []
        self.__post_triggers = {}

    def add_ignored_datasets(self, names):
        """Add h5 dataset name to ignore list"""
        self.__ignored_datasets.extend(names)

    def add_post_trigger(self, nodecls: BaseNode, callback: Callable):
        """Add post trigger function, to be called at the end of the collect process.
        Callbacks take only one argument: H5Importer instance."""
        triggers = self.__post_triggers.setdefault(nodecls, [])
        triggers.append(callback)

    def register(self, cls, is_generic=False):
        """Register node class.
        Generic classes are processed after specific classes (as a fallback solution)"""
        if is_generic:
            self.__generic_classes.append(cls)
        else:
            self.__thirdparty_classes.append(cls)

    def get(self, dset):
        """Return node class that matches h5 dataset"""
        for name in to_string(dset.name).split("/"):
            if name in self.__ignored_datasets:
                return None
        for cls in self.__thirdparty_classes + self.__generic_classes:
            if cls.match(dset):
                return cls
        if isinstance(dset, h5py.Group):
            return GroupNode
        return None

    def run_post_triggers(self, importer: H5Importer):
        """Run post-collect callbacks"""
        for node in importer.nodes:
            for nodecls, triggers in self.__post_triggers.items():
                if isinstance(node, nodecls):
                    for func in triggers:
                        func(node, importer)


NODE_FACTORY = NodeFactory()


class GroupNode(BaseNode):
    """Object representing a HDF5 group node"""

    @property
    def icon_name(self):
        """Icon name associated to node"""
        return "h5group.svg"

    def collect_children(self, node_dict: Dict):
        """Construct tree"""
        for dset in self.dset.values():
            child_cls = NODE_FACTORY.get(dset)
            if child_cls is not None:
                child = child_cls(self.h5file, dset.name)
                node_dict[child.id] = child
                self.children.append(child)
                if isinstance(child, GroupNode):
                    child.collect_children(node_dict)

    @property
    def text(self):
        """Return node textual representation"""
        return self.dset.name


class RootNode(GroupNode):
    """Object representing a HDF5 root node"""

    def __init__(self, h5file):
        super().__init__(h5file, "/")

    @property
    def icon_name(self):
        """Icon name associated to node"""
        return "h5file.svg"

    @property
    def name(self):
        """Return node name, constructed from dataset name"""
        return osp.basename(self.h5file.filename)

    @property
    def description(self):
        """Return node description"""
        return self.h5file.filename



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Application embedded test 2

CodraFT main window is simply hidden when closing application.
It is shown and raised above other windows when reopening application.
"""

from codraft.core.gui.main import CodraFTMainWindow
from codraft.tests import embedded1_unit

SHOW = True  # Show test in GUI-based test launcher


class HostWindow(embedded1_unit.BaseHostWindow):
    """Test main view"""

    def open_codraft(self):
        """Open CodraFT test"""
        if self.codraft is None:
            self.codraft = CodraFTMainWindow(console=False, hide_on_close=True)
            self.host.log("✨ Initialized CodraFT window")
            self.codraft.show()
        else:
            self.codraft.show()
            self.codraft.raise_()
        self.host.log("=> Shown CodraFT window")

    def close_codraft(self):
        """Close CodraFT window"""
        if self.codraft is not None:
            self.host.log("=> Closed CodraFT")
            self.codraft.close()


if __name__ == "__main__":
    embedded1_unit.test_embedded_feature(HostWindow)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT main window
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

import locale
import os
import os.path as osp
import platform
import sys
import time
import webbrowser
from typing import List

import numpy as np
import scipy.ndimage as spi
import scipy.signal as sps
from guidata import __version__ as guidata_ver
from guidata.configtools import get_icon, get_module_data_path, get_module_path
from guidata.qthelpers import add_actions, create_action, win32_fix_title_bar_background
from guidata.widgets.console import DockableConsole
from guiqwt import __version__ as guiqwt_ver
from guiqwt.builder import make
from guiqwt.plot import CurveWidget, ImageWidget
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW
from qtpy.compat import getopenfilenames, getsavefilename
from qwt import __version__ as qwt_ver

from codraft import __docurl__, __homeurl__, __supporturl__, __version__, env
from codraft.config import APP_DESC, APP_NAME, TEST_SEGFAULT_ERROR, Conf, _
from codraft.core.gui.actionhandler import ActionCategory
from codraft.core.gui.docks import DockablePlotWidget, DockableTabWidget
from codraft.core.gui.h5io import H5InputOutput
from codraft.core.gui.panel import ImagePanel, SignalPanel
from codraft.core.model.image import ImageParam
from codraft.core.model.signal import SignalParam
from codraft.env import execenv
from codraft.utils import dephash
from codraft.utils import qthelpers as qth
from codraft.widgets.instconfviewer import exec_codraft_installconfig_dialog
from codraft.widgets.logviewer import exec_codraft_logviewer_dialog
from codraft.widgets.status import MemoryStatus

DATAPATH = get_module_data_path("codraft", "data")


def get_htmlhelp():
    """Return HTML Help documentation link adapted to locale, if it exists"""
    if os.name == "nt":
        for suffix in ("_" + locale.getlocale()[0][:2], ""):
            path = osp.join(DATAPATH, f"CodraFT{suffix}.chm")
            if osp.isfile(path):
                return path
    return None


class AppProxy:
    """Proxy to CodraFT application: object used from the embedded console
    to access CodraFT internal objects"""

    def __init__(self, win):
        self.win = win
        self.s = self.win.signalpanel.objlist
        self.i = self.win.imagepanel.objlist


def is_frozen(module_name):
    """Test if module has been frozen (py2exe/cx_Freeze)"""
    datapath = get_module_path(module_name)
    parentdir = osp.normpath(osp.join(datapath, osp.pardir))
    return not osp.isfile(__file__) or osp.isfile(parentdir)  # library.zip


class CodraFTMainWindow(QW.QMainWindow):
    """CodraFT main window"""

    __instance = None

    @staticmethod
    def get_instance(console=None, hide_on_close=False):
        """Return singleton instance"""
        if CodraFTMainWindow.__instance is None:
            return CodraFTMainWindow(console, hide_on_close)
        return CodraFTMainWindow.__instance

    def __init__(self, console=None, hide_on_close=False):
        """Initialize main window"""
        CodraFTMainWindow.__instance = self
        super().__init__()
        win32_fix_title_bar_background(self)
        self.setObjectName(APP_NAME)
        self.setWindowIcon(get_icon("codraft.svg"))
        self.__restore_pos_and_size()

        self.hide_on_close = hide_on_close
        self.__old_size = None
        self.__memory_warning = False
        self.memorystatus = None

        self.console = None
        self.app_proxy = None
        self.signal_toolbar = None
        self.image_toolbar = None
        self.signalpanel = None
        self.imagepanel = None
        self.tabwidget = None
        self.signal_image_docks = None
        self.h5inputoutput = H5InputOutput(self)

        self.openh5_action = None
        self.saveh5_action = None
        self.browseh5_action = None
        self.quit_action = None

        self.file_menu = None
        self.edit_menu = None
        self.operation_menu = None
        self.processing_menu = None
        self.computing_menu = None
        self.view_menu = None
        self.help_menu = None

        self.__is_modified = None
        self.set_modified(False)

        # Setup actions and menus
        if console is None:
            console = Conf.console.enable.get(True)
        self.setup(console)

    @property
    def panels(self):
        """Return the tuple of implemented panels (signal, image)"""
        return (self.signalpanel, self.imagepanel)

    def __set_low_memory_state(self, state):
        """Set memory warning state"""
        self.__memory_warning = state

    def confirm_memory_state(self):  # pragma: no cover
        """Check memory warning state and eventually show a warning dialog"""
        if self.__memory_warning:
            threshold = Conf.main.available_memory_threshold.get()
            answer = QW.QMessageBox.critical(
                self,
                _("Warning"),
                _("Available memory is below %d MB.<br><br>Do you want to continue?")
                % threshold,
                QW.QMessageBox.Yes | QW.QMessageBox.No,
            )
            return answer == QW.QMessageBox.Yes
        return True

    def check_stable_release(self):  # pragma: no cover
        """Check if this is a stable release"""
        if __version__.replace(".", "").isdigit():
            # This is a stable release
            return
        if "b" in __version__:
            # This is a beta release
            rel = _(
                "This software is in the <b>beta stage</b> of its release cycle. "
                "The focus of beta testing is providing a feature complete "
                "software for users interested in trying new features before "
                "the final release. However, <u>beta software may not behave as "
                "expected and will probably have more bugs or performance issues "
                "than completed software</u>."
            )
        else:
            # This is an alpha release
            rel = _(
                "This software is in the <b>alpha stage</b> of its release cycle. "
                "The focus of alpha testing is providing an incomplete software "
                "for early testing of specific features by users. "
                "Please note that <u>alpha software was not thoroughly tested</u> "
                "by the developer before it is released."
            )
        txtlist = [
            f"<b>{APP_NAME}</b> v{__version__}:",
            "",
            _("<i>This is not a stable release.</i>"),
            "",
            rel,
        ]
        QW.QMessageBox.warning(self, APP_NAME, "<br>".join(txtlist), QW.QMessageBox.Ok)

    def check_dependencies(self):  # pragma: no cover
        """Check dependencies"""
        if is_frozen("codraft") or Conf.main.ignore_dependency_check.get(False):
            # No need to check dependencies if CodraFT has been frozen
            return
        try:
            state = dephash.check_dependencies_hash(DATAPATH)
            bad_deps = [name for name in state if not state[name]]
            if not bad_deps:
                # Everything is OK
                return
        except IOError:
            bad_deps = None
        txt0 = _("Non-compliant dependency:")
        if bad_deps is None or len(bad_deps) > 1:
            txt0 = _("Non-compliant dependencies:")
        if bad_deps is None:
            txtlist = [
                _("CodraFT has not yet been qualified on your operating system."),
            ]
        else:
            txtlist = [
                "<u>" + txt0 + "</u> " + ", ".join(bad_deps),
                "",
                "",
                _(
                    "At least one dependency does not comply with CodraFT "
                    "qualification standard reference (wrong dependency version "
                    "has been installed, or dependency source code has been "
                    "modified, or the application has not yet been qualified "
                    "on your operating system)."
                ),
            ]
        txtlist += [
            "",
            _(
                "This means that the application has not been officially qualified "
                "in this context and may not behave as expected."
            ),
            "",
            _(
                "Please click on the Ignore button "
                "to avoid showing this message at startup."
            ),
        ]
        txt = "<br>".join(txtlist)
        btn = QW.QMessageBox.information(
            self, APP_NAME, txt, QW.QMessageBox.Ok | QW.QMessageBox.Ignore
        )
        Conf.main.ignore_dependency_check.set(btn == QW.QMessageBox.Ignore)

    def check_for_previous_crash(self):  # pragma: no cover
        """Check for previous crash"""
        if execenv.unattended:
            self.show_log_viewer()
        elif Conf.main.faulthandler_log_available.get(
            False
        ) or Conf.main.traceback_log_available.get(False):
            txt = "<br>".join(
                [
                    _("Log files were generated during last session."),
                    "",
                    _("Do you want to see available log files?"),
                ]
            )
            btns = QW.QMessageBox.StandardButton.Yes | QW.QMessageBox.StandardButton.No
            choice = QW.QMessageBox.warning(self, APP_NAME, txt, btns)
            if choice == QW.QMessageBox.StandardButton.Yes:
                self.show_log_viewer()

    def take_screenshot(self, name):  # pragma: no cover
        """Take main window screenshot"""
        self.memorystatus.set_demo_mode(True)
        qth.grab_save_window(self, f"{name}")
        self.memorystatus.set_demo_mode(False)

    def take_menu_screenshots(self):  # pragma: no cover
        """Take menu screenshots"""
        for panel in self.panels:
            self.tabwidget.setCurrentWidget(panel)
            for name in (
                "file",
                "edit",
                "view",
                "operation",
                "processing",
                "computing",
                "help",
            ):
                menu = getattr(self, f"{name}_menu")
                menu.popup(self.pos())
                qth.grab_save_window(menu, f"{panel.objectName()}_{name}")
                menu.close()

    # ------GUI setup
    def __restore_pos_and_size(self):
        """Restore main window position and size from configuration"""
        pos = Conf.main.window_position.get(None)
        if pos is not None:
            posx, posy = pos
            self.move(QC.QPoint(posx, posy))
        size = Conf.main.window_size.get(None)
        if size is not None:
            width, height = size
            self.resize(QC.QSize(width, height))
        if pos is not None and size is not None:
            sgeo = self.screen().availableGeometry()
            out_inf = posx < -int(0.9 * width) or posy < -int(0.9 * height)
            out_sup = posx > int(0.9 * sgeo.width()) or posy > int(0.9 * sgeo.height())
            if len(QW.QApplication.screens()) == 1 and (out_inf or out_sup):
                #  Main window is offscreen
                posx = min(max(posx, 0), sgeo.width() - width)
                posy = min(max(posy, 0), sgeo.height() - height)
                self.move(QC.QPoint(posx, posy))

    def __save_pos_and_size(self):
        """Save main window position and size to configuration"""
        is_maximized = self.windowState() == QC.Qt.WindowMaximized
        Conf.main.window_maximized.set(is_maximized)
        if not is_maximized:
            size = self.size()
            Conf.main.window_size.set((size.width(), size.height()))
            pos = self.pos()
            Conf.main.window_position.set((pos.x(), pos.y()))

    def setup(self, console):
        """Setup main window"""
        self.statusBar().showMessage(_("Welcome to %s!") % APP_NAME, 5000)
        self.memorystatus = MemoryStatus(Conf.main.available_memory_threshold.get(500))
        self.memorystatus.SIG_MEMORY_ALARM.connect(self.__set_low_memory_state)
        self.statusBar().addPermanentWidget(self.memorystatus)
        self.__setup_commmon_actions()
        curvewidget = self.__add_signal_panel()
        imagewidget = self.__add_image_panel()
        self.__add_tabwidget(curvewidget, imagewidget)
        self.__add_menus()
        if console:
            self.__setup_console()
        # Update selection dependent actions
        self.__update_actions()
        self.signal_image_docks[0].raise_()
        # Restoring current tab from last session
        tab_idx = Conf.main.current_tab.get(None)
        if tab_idx is not None:
            self.tabwidget.setCurrentIndex(tab_idx)

    def __setup_commmon_actions(self):
        """Setup common actions"""
        self.openh5_action = create_action(
            self,
            _("Open HDF5 files..."),
            icon=get_icon("h5open.svg"),
            tip=_("Open one or several HDF5 files"),
            triggered=lambda checked=False: self.open_h5_files(import_all=True),
        )
        self.saveh5_action = create_action(
            self,
            _("Save to HDF5 file..."),
            icon=get_icon("h5save.svg"),
            tip=_("Save to HDF5 file"),
            triggered=self.save_to_h5_file,
        )
        self.browseh5_action = create_action(
            self,
            _("Browse HDF5 file..."),
            icon=get_icon("h5browser.svg"),
            tip=_("Browse an HDF5 file"),
            triggered=lambda checked=False: self.open_h5_files(import_all=None),
        )
        h5_toolbar = self.addToolBar(_("HDF5 I/O Toolbar"))
        add_actions(
            h5_toolbar, [self.openh5_action, self.saveh5_action, self.browseh5_action]
        )
        # Quit action for "File menu" (added when populating menu on demand)
        if self.hide_on_close:
            quit_text = _("Hide window")
            quit_tip = _("Hide CodraFT window")
        else:
            quit_text = _("Quit")
            quit_tip = _("Quit application")
        self.quit_action = create_action(
            self,
            quit_text,
            shortcut=QG.QKeySequence(QG.QKeySequence.Quit),
            icon=get_icon("libre-gui-close.svg"),
            tip=quit_tip,
            triggered=self.close,
        )

    def __add_signal_panel(self):
        """Setup signal toolbar, widgets and panel"""
        self.signal_toolbar = self.addToolBar(_("Signal Processing Toolbar"))
        curveplot_toolbar = self.addToolBar(_("Curve Plotting Toolbar"))
        curvewidget = DockablePlotWidget(self, CurveWidget, curveplot_toolbar)
        curveplot = curvewidget.get_plot()
        curveplot.add_item(make.legend("TR"))
        self.signalpanel = SignalPanel(
            self, curvewidget.plotwidget, self.signal_toolbar
        )
        self.signalpanel.SIG_STATUS_MESSAGE.connect(self.statusBar().showMessage)
        return curvewidget

    def __add_image_panel(self):
        """Setup image toolbar, widgets and panel"""
        self.image_toolbar = self.addToolBar(_("Image Processing Toolbar"))
        imagevis_toolbar = self.addToolBar(_("Image Visualization Toolbar"))
        imagewidget = DockablePlotWidget(self, ImageWidget, imagevis_toolbar)
        self.imagepanel = ImagePanel(self, imagewidget.plotwidget, self.image_toolbar)
        # -----------------------------------------------------------------------------
        # # Before eventually disabling the "peritem" mode by default, wait for the
        # # guiqwt bug to be fixed (peritem mode is not compatible with multiple image
        # # items):
        # for cspanel in (
        #     self.imagepanel.plotwidget.get_xcs_panel(),
        #     self.imagepanel.plotwidget.get_ycs_panel(),
        # ):
        #     cspanel.peritem_ac.setChecked(False)
        # -----------------------------------------------------------------------------
        self.imagepanel.SIG_STATUS_MESSAGE.connect(self.statusBar().showMessage)
        return imagewidget

    def switch_to_signal_panel(self):
        """Switch to signal panel"""
        self.tabwidget.setCurrentWidget(self.signalpanel)

    def switch_to_image_panel(self):
        """Switch to image panel"""
        self.tabwidget.setCurrentWidget(self.imagepanel)

    def __add_tabwidget(self, curvewidget, imagewidget):
        """Setup tabwidget with signals and images"""
        self.tabwidget = DockableTabWidget()
        self.tabwidget.setMaximumWidth(500)
        self.tabwidget.addTab(self.signalpanel, get_icon("signal.svg"), _("Signals"))
        self.tabwidget.addTab(self.imagepanel, get_icon("image.svg"), _("Images"))
        self.__add_dockwidget(self.tabwidget, _("Main panel"))
        curve_dock = self.__add_dockwidget(curvewidget, title=_("Curve panel"))
        image_dock = self.__add_dockwidget(imagewidget, title=_("Image panel"))
        self.tabifyDockWidget(curve_dock, image_dock)
        self.signal_image_docks = curve_dock, image_dock
        self.tabwidget.currentChanged.connect(self.__tab_index_changed)
        self.signalpanel.SIG_OBJECT_ADDED.connect(self.switch_to_signal_panel)
        self.imagepanel.SIG_OBJECT_ADDED.connect(self.switch_to_image_panel)
        for panel in self.panels:
            panel.SIG_OBJECT_ADDED.connect(self.set_modified)
            panel.SIG_OBJECT_REMOVED.connect(self.set_modified)

    def __add_menus(self):
        """Adding menus"""
        self.file_menu = self.menuBar().addMenu(_("File"))
        self.file_menu.aboutToShow.connect(self.__update_file_menu)
        self.edit_menu = self.menuBar().addMenu(_("&Edit"))
        self.operation_menu = self.menuBar().addMenu(_("Operations"))
        self.processing_menu = self.menuBar().addMenu(_("Processing"))
        self.computing_menu = self.menuBar().addMenu(_("Computing"))
        self.view_menu = self.menuBar().addMenu(_("&View"))
        self.view_menu.aboutToShow.connect(self.__update_view_menu)
        self.help_menu = self.menuBar().addMenu("?")
        for menu in (
            self.edit_menu,
            self.operation_menu,
            self.processing_menu,
            self.computing_menu,
        ):
            menu.aboutToShow.connect(self.__update_generic_menu)
        about_action = create_action(
            self,
            _("About..."),
            icon=get_icon("libre-gui-about.svg"),
            triggered=self.__about,
        )
        homepage_action = create_action(
            self,
            _("Project home page"),
            icon=get_icon("libre-gui-globe.svg"),
            triggered=lambda: webbrowser.open(__homeurl__),
        )
        issue_action = create_action(
            self,
            _("Bug report or feature request"),
            icon=get_icon("libre-gui-globe.svg"),
            triggered=lambda: webbrowser.open(__supporturl__),
        )
        onlinedoc_action = create_action(
            self,
            _("Online documentation"),
            icon=get_icon("libre-gui-help.svg"),
            triggered=lambda: webbrowser.open(__docurl__),
        )
        chmdoc_action = create_action(
            self,
            _("CHM documentation"),
            icon=get_icon("chm.svg"),
            triggered=lambda: os.startfile(get_htmlhelp()),
        )
        chmdoc_action.setVisible(get_htmlhelp() is not None)
        logv_action = create_action(
            self,
            _("Show log files..."),
            icon=get_icon("logs.svg"),
            triggered=self.show_log_viewer,
        )
        dep_action = create_action(
            self,
            _("About CodraFT installation") + "...",
            icon=get_icon("logs.svg"),
            triggered=lambda: exec_codraft_installconfig_dialog(self),
        )
        errtest_action = create_action(
            self, "Test segfault/Python error", triggered=self.test_segfault_error
        )
        errtest_action.setVisible(TEST_SEGFAULT_ERROR)
        about_action = create_action(
            self,
            _("About..."),
            icon=get_icon("libre-gui-about.svg"),
            triggered=self.__about,
        )
        add_actions(
            self.help_menu,
            (
                onlinedoc_action,
                chmdoc_action,
                None,
                errtest_action,
                logv_action,
                dep_action,
                None,
                homepage_action,
                issue_action,
                about_action,
            ),
        )

    def __setup_console(self):
        """Add an internal console"""
        self.app_proxy = AppProxy(self)
        ns = {
            "app": self.app_proxy,
            "np": np,
            "sps": sps,
            "spi": spi,
            "os": os,
            "sys": sys,
            "osp": osp,
            "time": time,
        }
        msg = (
            "Example: app.s[0] returns signal object #0\n"
            "Modules imported at startup: "
            "os, sys, os.path as osp, time, "
            "numpy as np, scipy.signal as sps, scipy.ndimage as spi"
        )
        debug = os.environ.get("DEBUG") == "1"
        self.console = DockableConsole(self, namespace=ns, message=msg, debug=debug)
        self.console.setMaximumBlockCount(Conf.console.max_line_count.get(5000))
        console_dock = self.__add_dockwidget(self.console, _("Console"))
        console_dock.hide()
        self.console.interpreter.widget_proxy.sig_new_prompt.connect(
            lambda txt: self.refresh_lists()
        )

    # ------GUI refresh
    def has_objects(self):
        """Return True if sig/ima panels have any object"""
        return sum([len(panel.objlist) for panel in self.panels]) > 0

    def set_modified(self, state=True):
        """Set mainwindow modified state"""
        state = state and self.has_objects()
        self.__is_modified = state
        self.setWindowTitle(APP_NAME + ("*" if state else ""))

    def __add_dockwidget(self, child, title):
        """Add QDockWidget and toggleViewAction"""
        dockwidget, location = child.create_dockwidget(title)
        self.addDockWidget(location, dockwidget)
        return dockwidget

    def refresh_lists(self):
        """Refresh signal/image lists"""
        for panel in self.panels:
            panel.objlist.refresh_list()

    def __update_actions(self):
        """Update selection dependent actions"""
        is_signal = self.tabwidget.currentWidget() is self.signalpanel
        panel = self.signalpanel if is_signal else self.imagepanel
        panel.selection_changed()
        self.signal_toolbar.setVisible(is_signal)
        self.image_toolbar.setVisible(not is_signal)

    def __tab_index_changed(self, index):
        """Switch from signal to image mode, or vice-versa"""
        dock = self.signal_image_docks[index]
        dock.raise_()
        self.__update_actions()

    def __update_generic_menu(self, menu=None):
        """Update menu before showing up -- Generic method"""
        if menu is None:
            menu = self.sender()
        menu.clear()
        panel = self.tabwidget.currentWidget()
        category = {
            self.file_menu: ActionCategory.FILE,
            self.edit_menu: ActionCategory.EDIT,
            self.view_menu: ActionCategory.VIEW,
            self.operation_menu: ActionCategory.OPERATION,
            self.processing_menu: ActionCategory.PROCESSING,
            self.computing_menu: ActionCategory.COMPUTING,
        }[menu]
        actions = panel.get_category_actions(category)
        add_actions(menu, actions)

    def __update_file_menu(self):
        """Update file menu before showing up"""
        self.saveh5_action.setEnabled(self.has_objects())
        self.__update_generic_menu(self.file_menu)
        add_actions(
            self.file_menu,
            [
                None,
                self.openh5_action,
                self.saveh5_action,
                self.browseh5_action,
                None,
                self.quit_action,
            ],
        )

    def __update_view_menu(self):
        """Update view menu before showing up"""
        self.__update_generic_menu(self.view_menu)
        add_actions(self.view_menu, [None] + self.createPopupMenu().actions())

    # ------Common features
    def reset_all(self):
        """Reset all application data"""
        for panel in self.panels:
            panel.remove_all_objects()

    @staticmethod
    def __check_h5file(filename, operation: str):
        """Check HDF5 filename"""
        filename = osp.abspath(osp.normpath(filename))
        bname = osp.basename(filename)
        if operation == "load" and not osp.isfile(filename):
            raise IOError(f'File not found "{bname}"')
        if not filename.endswith(".h5"):
            raise IOError(f'Invalid HDF5 file "{bname}"')
        Conf.main.base_dir.set(filename)
        return filename

    def save_to_h5_file(self, filename=None):
        """Save to a CodraFT HDF5 file"""
        if filename is None:
            basedir = Conf.main.base_dir.get()
            with qth.save_restore_stds():
                filters = f'{_("HDF5 files")} (*.h5)'
                filename, _filter = getsavefilename(self, _("Save"), basedir, filters)
            if not filename:
                return
        with qth.qt_try_loadsave_file(self.parent(), filename, "save"):
            filename = self.__check_h5file(filename, "save")
            self.h5inputoutput.save_file(filename)
            self.set_modified(False)

    def open_h5_files(
        self,
        h5files: List[str] = None,
        import_all: bool = None,
        reset_all: bool = None,
    ) -> None:
        """Open a CodraFT HDF5 file or import from any other HDF5 file

        :param h5files: HDF5 filenames (optionally with dataset name, separated by ":")
        :param import_all: Import all HDF5 file contents
        :param reset_all: Delete all CodraFT signals and images before importing data
        """
        if not self.confirm_memory_state():
            return
        if reset_all is None:
            reset_all = False
            if self.has_objects():
                answer = QW.QMessageBox.question(
                    self,
                    _("Warning"),
                    _(
                        "Do you want to remove all signals and images "
                        "before importing data from HDF5 files?"
                    ),
                    QW.QMessageBox.Yes | QW.QMessageBox.No,
                )
                if answer == QW.QMessageBox.Yes:
                    reset_all = True
        if h5files is None:
            basedir = Conf.main.base_dir.get()
            with qth.save_restore_stds():
                filters = f'{_("HDF5 files")} (*.h5)'
                h5files, _filter = getopenfilenames(self, _("Open"), basedir, filters)
        for fname_with_dset in h5files:
            if "," in fname_with_dset:
                filename, dsetname = fname_with_dset.split(",")
            else:
                filename, dsetname = fname_with_dset, None
            if import_all is None and dsetname is None:
                self.import_h5_file(filename, reset_all)
            else:
                with qth.qt_try_loadsave_file(self, filename, "load"):
                    filename = self.__check_h5file(filename, "load")
                    if dsetname is None:
                        self.h5inputoutput.open_file(filename, import_all, reset_all)
                    else:
                        self.h5inputoutput.import_dataset_from_file(filename, dsetname)
            reset_all = False

    def import_h5_file(self, filename: str, reset_all: bool = None) -> None:
        """Open CodraFT HDF5 browser to Import HDF5 file

        :param filename: HDF5 filename
        :param reset_all: Delete all CodraFT signals and images before importing data
        """
        with qth.qt_try_loadsave_file(self, filename, "load"):
            filename = self.__check_h5file(filename, "load")
            self.h5inputoutput.import_file(filename, False, reset_all)

    def add_object(self, obj, refresh=True):
        """Add object - signal or image"""
        if self.confirm_memory_state():
            if isinstance(obj, SignalParam):
                self.signalpanel.add_object(obj, refresh=refresh)
            elif isinstance(obj, ImageParam):
                self.imagepanel.add_object(obj, refresh=refresh)
            else:
                raise TypeError(f"Unsupported object type {type(obj)}")

    # ------?
    def __about(self):  # pragma: no cover
        """About dialog box"""
        self.check_stable_release()
        QW.QMessageBox.about(
            self,
            _("About ") + APP_NAME,
            f"""<b>{APP_NAME}</b> v{__version__}<br>{APP_DESC}<p>
              %s Pierre Raybaut
              <br>Copyright &copy; 2018-2022 CEA-CODRA
              <p>PythonQwt {qwt_ver}, guidata {guidata_ver},
              guiqwt {guiqwt_ver}<br>Python {platform.python_version()},
              Qt {QC.__version__}, PyQt {QC.PYQT_VERSION_STR}
               %s {platform.system()}"""
            % (_("Developped by"), _("on")),
        )

    def show_log_viewer(self):
        """Show error logs"""
        exec_codraft_logviewer_dialog(self)

    @staticmethod
    def test_segfault_error():
        """Generate errors (both fault and traceback)"""
        import ctypes  # pylint: disable=import-outside-toplevel

        ctypes.string_at(0)
        raise RuntimeError("!!! Testing RuntimeError !!!")

    def show(self):
        """Reimplement QMainWindow method"""
        super().show()
        if self.__old_size is not None:
            self.resize(self.__old_size)

    # ------Close window
    def closeEvent(self, event):
        """Reimplement QMainWindow method"""
        if self.hide_on_close:
            self.__old_size = self.size()
            self.hide()
        else:
            if not env.execenv.unattended and self.__is_modified:
                answer = QW.QMessageBox.warning(
                    self,
                    _("Quit"),
                    _(
                        "Do you want to save all signals and images "
                        "to an HDF5 file before quitting CodraFT?"
                    ),
                    QW.QMessageBox.Yes | QW.QMessageBox.No | QW.QMessageBox.Cancel,
                )
                if answer == QW.QMessageBox.Yes:
                    self.save_to_h5_file()
                    if self.__is_modified:
                        event.ignore()
                        return
                elif answer == QW.QMessageBox.Cancel:
                    event.ignore()
                    return
            if self.console is not None:
                try:
                    self.console.close()
                except RuntimeError:
                    # TODO: [P3] Investigate further why the following error occurs when
                    # restarting the mainwindow (this is *not* a production case):
                    # "RuntimeError: wrapped C/C++ object of type DockableConsole
                    #  has been deleted".
                    # Another solution to avoid this error would be to really restart
                    # the application (run each unit test in a separate process), but
                    # it would represent too much effort for an error occuring in test
                    # configurations only.
                    pass
            self.reset_all()
            self.__save_pos_and_size()

            # Saving current tab for next session
            Conf.main.current_tab.set(self.tabwidget.currentIndex())

            event.accept()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
codraft.config
--------------

The `config` module handles `codraft` configuration
(options, images and icons).
"""

import os
import os.path as osp

from guidata import configtools
from guiqwt.config import CONF as GUIQWT_CONF

from codraft.utils import conf, tests

_ = configtools.get_translation("codraft")

CONF_VERSION = "1.0.0"
APP_NAME = "CodraFT"
APP_DESC = _(
    """CodraFT (<b>Codra</b> <b>F</b>iltering <b>T</b>ool) is a
generic signal and image processing software based on Python and Qt"""
)
APP_PATH = osp.dirname(__file__)
DEBUG = len(os.environ.get("DEBUG", "")) > 0
if DEBUG:
    print("*** DEBUG mode *** [Reset configuration file, do not redirect std I/O]")
TEST_SEGFAULT_ERROR = len(os.environ.get("TEST_SEGFAULT_ERROR", "")) > 0
if TEST_SEGFAULT_ERROR:
    print('*** TEST_SEGFAULT_ERROR mode *** [Enabling test action in "?" menu]')
DATETIME_FORMAT = "%d/%m/%Y - %H:%M:%S"

configtools.add_image_module_path("codraft", osp.join("data", "logo"))
configtools.add_image_module_path("codraft", osp.join("data", "icons"))


class MainSection(conf.Section, metaclass=conf.SectionMeta):
    """Class defining the main configuration section structure.
    Each class attribute is an option (metaclass is automatically affecting
    option names in .INI file based on class attribute names)."""

    traceback_log_path = conf.ConfigPathOption()
    traceback_log_available = conf.Option()
    faulthandler_enabled = conf.Option()
    faulthandler_log_path = conf.ConfigPathOption()
    faulthandler_log_available = conf.Option()
    window_maximized = conf.Option()
    window_position = conf.Option()
    window_size = conf.Option()
    base_dir = conf.WorkingDirOption()
    available_memory_threshold = conf.Option()
    ignore_dependency_check = conf.Option()
    current_tab = conf.Option()


class ConsoleSection(conf.Section, metaclass=conf.SectionMeta):
    """Classs defining the console configuration section structure.
    Each class attribute is an option (metaclass is automatically affecting
    option names in .INI file based on class attribute names)."""

    enable = conf.Option()
    max_line_count = conf.Option()


class IOSection(conf.Section, metaclass=conf.SectionMeta):
    """Class defining the I/O configuration section structure.
    Each class attribute is an option (metaclass is automatically affecting
    option names in .INI file based on class attribute names)."""

    h5_fname_in_title = conf.Option()
    h5_fullpath_in_title = conf.Option()


class ProcSection(conf.Section, metaclass=conf.SectionMeta):
    """Class defining the Processing configuration section structure.
    Each class attribute is an option (metaclass is automatically affecting
    option names in .INI file based on class attribute names)."""

    extract_roi_singleobj = conf.Option()


class ViewSection(conf.Section, metaclass=conf.SectionMeta):
    """Class defining the view configuration section structure.
    Each class attribute is an option (metaclass is automatically affecting
    option names in .INI file based on class attribute names)."""

    # String formatting for shape legends
    sig_format = conf.Option()
    ima_format = conf.Option()

    show_label = conf.Option()
    show_contrast = conf.Option()

    # If True, images are shown with the same LUT range as the first selected image
    ima_ref_lut_range = conf.Option()

    # Default visualization settings at item creation
    # (e.g. see `ImageParam.make_item` in codraft/core/model/image.py)
    ima_eliminate_outliers = conf.Option()

    # Default visualization settings, persisted in object metadata
    # (e.g. see `create_image` in codraft/core/model/image.py)
    ima_def_colormap = conf.Option()
    ima_def_interpolation = conf.Option()


# Usage (example): Conf.console.enable.get(True)
class Conf(conf.Configuration, metaclass=conf.ConfMeta):
    """Class defining CodraFT configuration structure.
    Each class attribute is a section (metaclass is automatically affecting
    section names in .INI file based on class attribute names)."""

    main = MainSection()
    console = ConsoleSection()
    view = ViewSection()
    proc = ProcSection()
    io = IOSection()


def get_old_log_fname(fname):
    """Return old log fname from current log fname"""
    return osp.splitext(fname)[0] + ".1.log"


def initialize():
    """Initialize application configuration"""
    Conf.initialize(APP_NAME, CONF_VERSION, load=not DEBUG)
    Conf.main.traceback_log_path.set(f".{APP_NAME}_traceback.log")
    Conf.main.faulthandler_log_path.set(f".{APP_NAME}_faulthandler.log")


def reset():
    """Reset application configuration"""
    Conf.reset()
    initialize()


initialize()
tests.add_test_module_path("codraft", osp.join("data", "tests"))


GUIQWT_DEFAULTS = {
    "plot": {
        # "antialiasing": False,
        # "title/font/size": 12,
        # "title/font/bold": False,
        # "marker/curve/text/font/size": 8,
        # "marker/curve/text/font/family": "default",
        # "marker/curve/text/font/bold": False,
        # "marker/curve/text/font/italic": False,
        "marker/curve/text/textcolor": "black",
        # "marker/curve/text/background_color": "#ffffff",
        # "marker/curve/text/background_alpha": 0.8,
        # "marker/cross/text/font/family": "default",
        # "marker/cross/text/font/size": 8,
        # "marker/cross/text/font/bold": False,
        # "marker/cross/text/font/italic": False,
        "marker/cross/text/textcolor": "black",
        # "marker/cross/text/background_color": "#ffffff",
        "marker/cross/text/background_alpha": 0.7,
        # "marker/cross/line/style": "DashLine",
        # "marker/cross/line/color": "yellow",
        # "marker/cross/line/width": 1,
        # "marker/cursor/text/font/size": 8,
        # "marker/cursor/text/font/family": "default",
        # "marker/cursor/text/font/bold": False,
        # "marker/cursor/text/font/italic": False,
        # "marker/cursor/text/textcolor": "#ff9393",
        # "marker/cursor/text/background_color": "#ffffff",
        # "marker/cursor/text/background_alpha": 0.8,
        "shape/drag/symbol/marker": "NoSymbol",
        "shape/mask/symbol/size": 5,
        "shape/mask/sel_symbol/size": 8,
    },
}

GUIQWT_CONF.update_defaults(GUIQWT_DEFAULTS)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Unit test scenario Signal 01

Testing the following:
  - Add signal object at startup
  - Create signal (random type)
  - Sum two signals
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code

from codraft.config import _
from codraft.core.gui.main import CodraFTMainWindow
from codraft.core.gui.processor.base import (
    ClipParam,
    GaussianParam,
    MovingAverageParam,
    MovingMedianParam,
    ThresholdParam,
)
from codraft.core.gui.processor.signal import (
    FWHMParam,
    NormalizeParam,
    PeakDetectionParam,
    PolynomialFitParam,
    XYCalibrateParam,
)
from codraft.core.model.base import UniformRandomParam
from codraft.core.model.signal import (
    GaussLorentzVoigtParam,
    SignalParamNew,
    SignalTypes,
    create_signal_from_param,
    new_signal_param,
)
from codraft.tests import codraft_app_context
from codraft.tests.data import create_test_signal1
from codraft.tests.newobject_unit import iterate_signal_creation
from codraft.widgets import fitdialog

SHOW = True  # Show test in GUI-based test launcher


def test_compute_11_operations(panel, index):
    """Test compute_11 type operations on a signal or image

    Requires that one signal or image has been added at index."""
    assert len(panel.objlist) >= index - 1
    panel.objlist.select_rows((index,))
    panel.processor.compute_gaussian(GaussianParam())
    panel.processor.compute_moving_average(MovingAverageParam())
    panel.processor.compute_moving_median(MovingMedianParam())
    panel.processor.compute_wiener()
    panel.processor.compute_fft()
    panel.processor.compute_ifft()
    panel.processor.compute_abs()
    panel.processor.compute_log10()
    panel.processor.swap_axes()
    panel.processor.swap_axes()


def test_common_operations(panel):
    """Test operations common to signal/image

    Requires that two (and only two) signals/images are created/added to panel

    First signal/image is supposed to be always the same (reference)
    Second signal/image is the tested object
    """
    assert len(panel.objlist) == 2

    panel.duplicate_object()
    panel.objlist.select_rows((1, 2))
    panel.processor.compute_difference(False)
    panel.remove_object()
    panel.objlist.select_rows((1, 2))
    panel.processor.compute_difference(True)
    panel.delete_metadata()
    panel.objlist.select_rows((2, 3))
    panel.remove_object()

    panel.objlist.select_rows((0, 1))
    panel.processor.compute_sum()
    panel.objlist.select_rows((0, 1))
    panel.processor.compute_sum()
    panel.objlist.select_rows((0, 1))
    panel.processor.compute_product()

    obj = panel.objlist[-1]
    param = ThresholdParam()
    param.value = (obj.data.max() - obj.data.min()) * 0.2 + obj.data.min()
    panel.processor.compute_threshold(param)
    param = ClipParam()  # Clipping before division...
    param.value = (obj.data.max() - obj.data.min()) * 0.8 + obj.data.min()
    panel.processor.compute_clip(param)

    panel.objlist.select_rows((2, 6))
    panel.processor.compute_division()
    panel.objlist.select_rows((0, 1, 2))
    panel.processor.compute_average()

    test_compute_11_operations(panel, 1)


def test_signal_features(win: CodraFTMainWindow, data_size: int = 500) -> None:
    """Testing signal features"""
    panel = win.signalpanel
    win.switch_to_signal_panel()

    for signal in iterate_signal_creation(data_size, non_zero=True):
        panel.add_object(create_test_signal1(data_size))
        panel.add_object(signal)
        test_common_operations(panel)
        panel.remove_all_objects()

    sig1 = create_test_signal1(data_size)
    win.add_object(sig1)

    # Add new signal based on s0
    panel.objlist.set_current_row(0)
    newparam = new_signal_param(_("Random function"), stype=SignalTypes.UNIFORMRANDOM)
    addparam = UniformRandomParam()
    addparam.vmin = 0
    addparam.vmax = sig1.y.max() * 0.2
    panel.new_object(newparam, addparam=addparam, edit=False)

    test_common_operations(panel)

    win.add_object(create_test_signal1(data_size))

    param = NormalizeParam()
    for _name, method in param.methods:
        param.method = method
        panel.processor.normalize(param)

    param = XYCalibrateParam()
    param.a, param.b = 1.2, 0.1
    panel.processor.calibrate(param)

    panel.processor.compute_derivative()
    panel.processor.compute_integral()

    param = PeakDetectionParam()
    panel.processor.detect_peaks(param)

    panel.processor.compute_multigaussianfit()

    panel.objlist.set_current_row(-3)
    sig = panel.objlist.get_sel_object()
    i1 = data_size // 10
    i2 = len(sig.y) - i1
    panel.processor.extract_roi([[i1, i2]])

    param = PolynomialFitParam()
    panel.processor.compute_polyfit(param)

    panel.processor.compute_fit(_("Gaussian fit"), fitdialog.gaussianfit)
    panel.processor.compute_fit(_("Lorentzian fit"), fitdialog.lorentzianfit)
    panel.processor.compute_fit(_("Voigt fit"), fitdialog.voigtfit)

    newparam = SignalParamNew()
    newparam.title = _("Gaussian")
    newparam.type = SignalTypes.GAUSS
    sig = create_signal_from_param(newparam, GaussLorentzVoigtParam(), edit=False)
    panel.add_object(sig)

    param = FWHMParam()
    for fittype, _name in param.fittypes:
        param.fittype = fittype
        panel.processor.compute_fwhm(param)
    panel.processor.compute_fw1e2()


def test():
    """Run signal unit test scenario 1"""
    with codraft_app_context(save=True) as win:
        test_signal_features(win)
        win.signalpanel.open_separate_view((0, 1, 2, 6))


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Log viewer test
"""

import os.path as osp
import subprocess
import sys

from codraft.app import run
from codraft.env import execenv

SHOW = True  # Show test in GUI-based test launcher


def exec_script(path):
    """Run test script"""
    command = [sys.executable, '"' + path + '"']
    stderr = subprocess.DEVNULL if execenv.unattended else None
    with subprocess.Popen(" ".join(command), shell=True, stderr=stderr) as proc:
        proc.wait()


if __name__ == "__main__":
    exec_script(osp.join(osp.dirname(__file__), "logview_error.py"))
    run()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Object (signal/image) list widgets
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

import re
from typing import Tuple

from qtpy import QtCore as QC
from qtpy import QtWidgets as QW

from codraft.utils.qthelpers import block_signals


class SimpleObjectList(QW.QListWidget):
    """Base object handling panel list widget, object (sig/ima) lists"""

    SIG_ITEM_DOUBLECLICKED = QC.Signal(int)
    SIG_CONTEXT_MENU = QC.Signal(QC.QPoint)

    def __init__(self, panel, parent=None):
        parent = panel if parent is None else parent
        super().__init__(parent)
        self.panel = panel
        self.prefix = panel.PREFIX
        self.setAlternatingRowColors(True)
        self._objects = []  # signals or images
        self.itemDoubleClicked.connect(self.item_double_clicked)

    def init_from(self, objlist):
        """Init from another SimpleObjectList, without making copies of objects"""
        self._objects = objlist.get_objects()
        self.refresh_list()
        self.setCurrentRow(objlist.currentRow())

    def get_objects(self):
        """Get all objects"""
        return self._objects

    def set_current_row(self, row, extend=False, refresh=True):
        """Set list widget current row"""
        if row < 0:
            row += self.count()
        if extend:
            command = QC.QItemSelectionModel.Select
        else:
            command = QC.QItemSelectionModel.ClearAndSelect
        with block_signals(widget=self, enable=not refresh):
            self.setCurrentRow(row, command)

    def refresh_list(self, new_current_row=None):
        """
        Refresh object list

        :param new_current_row: New row (if None, new current row is unchanged)
        """
        row = self.currentRow()
        if new_current_row is not None:
            row = new_current_row
        self.clear()
        for idx, obj in enumerate(self._objects):
            item = QW.QListWidgetItem(f"{self.prefix}{idx:03d}: {obj.title}", self)
            item.setToolTip(obj.metadata_to_html())
            self.addItem(item)
        if row < self.count():
            self.set_current_row(row)

    def item_double_clicked(self, listwidgetitem):
        """Item was double-clicked: open a pop-up plot dialog"""
        self.SIG_ITEM_DOUBLECLICKED.emit(self.row(listwidgetitem))

    def contextMenuEvent(self, event):  # pylint: disable=C0103
        """Override Qt method"""
        self.SIG_CONTEXT_MENU.emit(event.globalPos())


class GetObjectDialog(QW.QDialog):
    """Get object dialog box"""

    def __init__(self, parent, panel, title):
        super().__init__(parent)
        self.setWindowTitle(title)
        self.setLayout(QW.QVBoxLayout())

        self.objlist = SimpleObjectList(panel, parent=parent)
        self.objlist.init_from(panel.objlist)
        self.objlist.SIG_ITEM_DOUBLECLICKED.connect(lambda row: self.accept())
        self.layout().addWidget(self.objlist)

        bbox = QW.QDialogButtonBox(QW.QDialogButtonBox.Ok | QW.QDialogButtonBox.Cancel)
        bbox.accepted.connect(self.accept)
        bbox.rejected.connect(self.reject)
        bbox.button(QW.QDialogButtonBox.Ok).setEnabled(self.objlist.count() > 0)
        self.layout().addSpacing(10)
        self.layout().addWidget(bbox)

    def get_object(self):
        """Return current object"""
        return self.objlist.get_objects()[self.objlist.currentRow()]


class ObjectList(SimpleObjectList):
    """Object handling panel list widget, object (sig/ima) lists"""

    def __init__(self, panel):
        super().__init__(panel)
        self.setSelectionMode(QW.QListWidget.ExtendedSelection)

    def __len__(self):
        """Return number of objects"""
        return len(self._objects)

    def __getitem__(self, row):
        """Return object at row"""
        return self._objects[row]

    def __setitem__(self, row, obj):
        """Set object at row"""
        self._objects[row] = obj

    def __contains__(self, obj):
        """Return True if list contain obj"""
        return obj in self._objects

    def get_row(self, obj):
        """Return row associated to object obj"""
        return self._objects.index(obj)

    def __fix_obj_titles(self, row: int, sign: int) -> None:
        """Fix all object titles before adding (sign==1) or removing (sign==-1)
        an object at row index"""
        pfx = self.prefix
        oname = f"{pfx}%03d"
        for obj in self:
            for match in re.finditer(pfx + "[0-9]{3}", obj.title):
                before = match.group()
                i_match = int(before[1:])
                if sign == -1 and i_match == row:
                    after = f"{pfx}xxx"
                elif (sign == -1 and i_match > row) or (sign == 1 and i_match >= row):
                    after = oname % (i_match + sign)
                else:
                    continue
                obj.title = obj.title.replace(before, after)

    def __delitem__(self, row):
        """Del object at row"""
        self.__fix_obj_titles(row, -1)
        self._objects.pop(row)

    def __iter__(self):
        """Return an iterator over objects"""
        yield from self._objects

    def get_sel_object(self, position=0):
        """
        Return currently selected object

        :param int position: Position in selection list (0 means first, -1 means last)
        :return: Current object or None if there is no selection
        """
        rows = self.get_selected_rows()
        if rows:
            return self[rows[position]]
        return None

    def get_sel_objects(self):
        """Return selected objects"""
        return [self[row] for row in self.get_selected_rows()]

    def append(self, obj):
        """Append object"""
        self._objects.append(obj)

    def insert(self, row, obj):
        """Insert object at row index"""
        self.__fix_obj_titles(row, 1)
        self._objects.insert(row, obj)

    def remove_all(self):
        """Remove all objects"""
        self._objects = []

    def select_rows(self, rows: Tuple):
        """Select multiple list widget rows"""
        for index, row in enumerate(sorted(rows)):
            self.set_current_row(row, extend=index != 0, refresh=row == len(rows) - 1)

    def select_all_rows(self):
        """Select all widget rows"""
        self.selectAll()

    def get_selected_rows(self):
        """Return selected rows"""
        return [index.row() for index in self.selectionModel().selectedRows()]
